# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains FreeCAD commands for the BIM workbench"""

import math
import os

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_ProjectManager:

    def GetResources(self):

        return {
            "Pixmap": "BIM_ProjectManager",
            "MenuText": QT_TRANSLATE_NOOP("BIM_ProjectManager", "Setup Project"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_ProjectManager", "Creates or manages a BIM project"),
        }

    def Activated(self):

        # only raise the dialog if it is already open
        if getattr(self, "form", None):
            self.form.raise_()
            return

        import FreeCADGui
        import ArchBuildingPart
        from PySide import QtGui

        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogProjectManager.ui")
        self.project = None
        self.site = None
        self.building = None

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        # set things up
        self.form.buildingUse.addItems(ArchBuildingPart.BuildingTypes)
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_ProjectManager.svg"))
        self.form.buttonAdd.clicked.connect(self.addGroup)
        self.form.buttonDel.clicked.connect(self.delGroup)
        self.form.buttonSave.clicked.connect(self.savePreset)
        self.form.presets.currentIndexChanged.connect(self.getPreset)
        self.form.buttonOK.clicked.connect(self.accept)
        self.form.buttonCancel.clicked.connect(self.reject)
        self.form.rejected.connect(self.reject)
        self.fillPresets()

        # Detect existing objects
        sel = FreeCADGui.Selection.getSelection()
        doc = FreeCAD.ActiveDocument
        if doc:
            if len(sel) == 1:
                if hasattr(sel[0], "Proxy") and hasattr(sel[0].Proxy, "ifcfile"):
                    # case 1: a project is selected
                    self.project = sel[0]
                    self.form.groupNewProject.setEnabled(False)
            if hasattr(doc, "Proxy"):
                # case 2: the actuve document is a project
                if hasattr(doc.Proxy, "ifcfile"):
                    self.project = doc
                    self.form.groupNewProject.setEnabled(False)
            if self.project:
                from nativeifc import ifc_tools

                sites = ifc_tools.get_children(self.project, ifctype="IfcSite")
                sites = list(filter(None, [ifc_tools.get_object(s) for s in sites]))
                self.form.projectName.setText(self.project.Label)
            else:
                sites = [o for o in doc.Objects if getattr(o, "IfcType", "") == "Site"]
            if sites:
                self.site = sites[0]
                self.form.siteName.setText(self.site.Label)
                if hasattr(self.site, "Address"):
                    self.form.siteAddress.setText(self.site.Address)
                elif hasattr(self.site, "SiteAddress"):
                    self.form.siteAddress.setText(self.site.SiteAddress)
                if hasattr(self.site, "Longitude"):
                    self.form.siteLongitude.setValue(self.site.Longitude)
                elif hasattr(self.site, "RefLongitude"):
                    self.form.siteLongitude.setValue(self.site.RefLongitude)
                if hasattr(self.site, "Latitude"):
                    self.form.siteLatitude.setValue(self.site.Latitude)
                elif hasattr(self.site, "RefLatitude"):
                    self.form.siteLatitude.setValue(self.site.RefLatitude)
                if hasattr(self.site, "Elevation"):
                    self.form.siteElevation.setText(self.site.Elevation.UserString)
                elif hasattr(self.site, "RefElevation"):
                    self.form.siteElevation.setText(self.site.RefElevation.UserString)
                if hasattr(self.site, "Declination"):
                    self.form.siteElevation.setText(str(self.site.Declination))
            buildings = []
            if self.site and self.project:
                from nativeifc import ifc_tools

                buildings = ifc_tools.get_children(self.site, ifctype="IfcBuilding")
                buildings = list(filter(None, [ifc_tools.get_object(b) for b in buildings]))
            if not buildings:
                buildings = [o for o in doc.Objects if getattr(o, "IfcType", "") == "Building"]
            if buildings:
                self.building = buildings[0]
                self.form.buildingName.setText(self.building.Label)
            levels = []
            if self.building and self.project:
                from nativeifc import ifc_tools

                levels = ifc_tools.get_children(self.building, ifctype="IfcBuildingStorey")
                levels = list(filter(None, [ifc_tools.get_object(l) for l in levels]))
            if not levels:
                levels = [o for o in doc.Objects if getattr(o, "IfcType", "") == "Building Storey"]
            if levels:
                self.form.countLevels.setValue(len(levels))

        # show dialog
        self.form.show()

    def reject(self):

        self.form.hide()
        del self.form
        return True

    def accept(self):

        import Arch
        import Draft
        import FreeCADGui
        import Part
        from draftutils import params

        vaxes = []
        haxes = []
        outline = None
        outtext = None
        human = None
        grp = None

        if self.form.groupNewProject.isChecked():
            self.project = None
            self.site = None
            self.building = None

        # reading form values
        buildingWidth = FreeCAD.Units.Quantity(self.form.buildingWidth.text()).Value
        buildingLength = FreeCAD.Units.Quantity(self.form.buildingLength.text()).Value
        distVAxes = FreeCAD.Units.Quantity(self.form.distVAxes.text()).Value
        distHAxes = FreeCAD.Units.Quantity(self.form.distHAxes.text()).Value
        levelHeight = FreeCAD.Units.Quantity(self.form.levelHeight.text()).Value
        color = self.form.lineColor.property("color").getRgbF()[:3]

        # Document creation
        doc = FreeCAD.ActiveDocument
        if self.form.groupNewProject.isChecked():
            if (
                self.form.radioNative1.isChecked()
                or self.form.radioNative3.isChecked()
                or (self.form.radioNative2.isChecked() and doc is None)
            ):
                doc = FreeCAD.newDocument()
                if self.form.projectName.text():
                    doc.Label = self.form.projectName.text()
        if doc:
            FreeCAD.setActiveDocument(doc.Name)

        # Project creation
        if self.form.groupNewProject.isChecked():
            from nativeifc import ifc_tools

            if self.form.radioNative2.isChecked():
                self.project = ifc_tools.create_document_object(doc, silent=True)
                if self.form.projectName.text():
                    self.project.Label = self.form.projectName.text()
            elif self.form.radioNative3.isChecked():
                self.project = ifc_tools.convert_document(doc, silent=True)

        # Human
        human = None
        if self.form.addHumanFigure.isChecked():
            from draftguitools import gui_trackers

            pts = gui_trackers.gridTracker.get_human_figure(None)
            human = FreeCAD.ActiveDocument.addObject("Part::Feature", "Human")
            human.Shape = Part.makePolygon(pts)
            human.Placement.move(FreeCAD.Vector(500, 500, 0))

        # Site creation or edition
        outline = None
        if self.form.groupSite.isChecked():
            if not self.site:
                self.site = Arch.makeSite()
                if self.project:
                    from nativeifc import ifc_tools

                    self.site = ifc_tools.aggregate(self.site, self.project)
            self.site.Label = self.form.siteName.text()
            if hasattr(self.site, "Address"):
                self.site.Address = self.form.siteAddress.text()
            elif hasattr(self.site, "SiteAddress"):
                self.site.SiteAddress = self.form.siteAddress.text()
            if hasattr(self.site, "Longitude"):
                self.site.Longitude = self.form.siteLongitude.value()
            elif hasattr(self.site, "RefLongitude"):
                self.site.RefLongitude = self.form.siteLongitude.value()
            if hasattr(self.site, "Latitude"):
                self.site.Latitude = self.form.siteLatitude.value()
            elif hasattr(self.site, "RefLatitude"):
                self.site.RefLatitude = self.form.siteLatitude.value()
            if hasattr(self.site, "NorthDeviation"):
                self.site.NorthDeviation = self.form.siteDeviation.value()
            elif hasattr(self.site, "Declination"):
                self.site.Declination = self.form.siteDeviation.value()
            elev = FreeCAD.Units.Quantity(self.form.siteElevation.text()).Value
            if hasattr(self.site, "Elevation"):
                self.site.Elevation = elev
            elif hasattr(self.site, "RefElevation"):
                self.site.RefElevation = elev

        # Building creation or edition
        if self.form.groupBuilding.isChecked():
            if not self.building:
                self.building = Arch.makeBuilding()
                if self.project:
                    from nativeifc import ifc_tools

                    if self.site:
                        self.building = ifc_tools.aggregate(self.building, self.site)
                    else:
                        self.building = ifc_tools.aggregate(self.building, self.project)
                elif self.site:
                    self.site.Group = [self.building]
            self.building.Label = self.form.buildingName.text()
            if hasattr(self.building, "BuildingType"):
                self.building.BuildingType = self.form.buildingUse.currentText()

            # Detecting existing contents
            if self.building:
                grp = [o for o in self.building.Group if o.Name.startswith("BuildingLayout")]
                if grp:
                    grp = grp[0]
            if grp:
                axes = [o for o in grp.Group if o.Name.startswith("Axis")]
                if axes:
                    for ax in axes:
                        if round(math.degrees(ax.Placement.Rotation.Angle), 0) in [0, 180, 360]:
                            vaxes.append(ax)
                        elif round(math.degrees(ax.Placement.Rotation.Angle), 0) in [90, 270]:
                            haxes.append(ax)
                outline = [o for o in grp.Group if o.Name.startswith("Rectangle")]
                if outline:
                    outline = outline[0]
                human = [o for o in grp.Group if o.Name.startswith("Human")]
                if human:
                    human = human[0]
            else:
                if (
                    self.form.addHumanFigure.isChecked()
                    or self.form.countVAxes.value()
                    or self.form.countHAxes.value()
                    or (buildingWidth and buildingLength)
                ):
                    grp = doc.addObject("App::DocumentObjectGroup", "BuildingLayout")
                    grp.Label = translate("BIM", "Building Layout")
                    if self.building:
                        if hasattr(self.building, "addObject"):
                            self.building.addObject(grp)

            # Human figure
            if self.form.addHumanFigure.isChecked():
                if not human:
                    # TODO embed this
                    humanpath = os.path.join(
                        os.path.dirname(__file__), "geometry", "human figure.brep"
                    )
                    if os.path.exists(humanpath):
                        humanshape = Part.Shape()
                        humanshape.importBrep(humanpath)
                        human = FreeCAD.ActiveDocument.addObject("Part::Feature", "Human")
                        human.Shape = humanshape
                        human.Placement.move(FreeCAD.Vector(500, 500, 0))
                    if human:
                        grp.addObject(human)
                    # TODO: nativeifc

            # Outline
            if buildingWidth and buildingLength:
                if not outline:
                    outline = Draft.makeRectangle(buildingLength, buildingWidth, face=False)
                    outline.Label = translate("BIM", "Building Outline")
                    outline.ViewObject.DrawStyle = "Dashed"
                    grp.addObject(outline)
                outline.ViewObject.LineColor = color
                outline.ViewObject.LineWidth = self.form.lineWidth.value() * 2
                outline.Length = buildingLength
                outline.Height = buildingWidth

            # Label
            if self.form.buildingName.text():
                buildingname = self.form.buildingName.text()
                outtext = Draft.make_text(
                    [buildingname],
                    FreeCAD.Vector(
                        params.get_param("textheight") * 0.3,
                        -params.get_param("textheight") * 1.43,
                        0,
                    ),
                )
                outtext.Label = translate("BIM", "Building Label")
                outtext.ViewObject.TextColor = color
                grp.addObject(outtext)

            # Axes
            axisV = None
            if self.form.countVAxes.value() and distVAxes:
                axisV = Arch.makeAxis(
                    num=self.form.countVAxes.value(), size=distVAxes, name="vaxis"
                )
                axisV.Label = translate("BIM", "Vertical Axes")
                axisV.ViewObject.BubblePosition = "Both"
                axisV.ViewObject.LineWidth = self.form.lineWidth.value()
                axisV.ViewObject.FontSize = params.get_param("textheight")
                axisV.ViewObject.BubbleSize = params.get_param("textheight") * 1.43
                axisV.ViewObject.LineColor = color
            axisH = None
            if self.form.countHAxes.value() and distHAxes:
                axisH = Arch.makeAxis(
                    num=self.form.countHAxes.value(), size=distHAxes, name="haxis"
                )
                axisH.Label = translate("BIM", "Horizontal Axes")
                axisH.ViewObject.BubblePosition = "Both"
                axisH.ViewObject.NumberingStyle = "A,B,C"
                axisH.ViewObject.LineWidth = self.form.lineWidth.value()
                axisH.ViewObject.FontSize = params.get_param("textheight")
                axisH.ViewObject.BubbleSize = params.get_param("textheight") * 1.43
                axisH.Placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 90)
                axisH.ViewObject.LineColor = color
            if axisV and axisH:
                axisG = Arch.makeAxisSystem([axisH, axisV])
                axisG.Label = translate("BIM", "Axes")
                grp.addObject(axisG)
            else:
                if axisV:
                    grp.addObject(axisV)
                if axisH:
                    grp.addObject(axisH)
            if self.form.countLevels.value() and levelHeight:
                h = 0
                alabels = []
                groups = []
                for i in range(self.form.groupsList.count()):
                    groups.append(self.form.groupsList.item(i).text())
                for i in range(self.form.countLevels.value()):
                    lev = Arch.makeFloor()
                    lev.Label = translate("BIM", "Level") + " " + str(i)
                    alabels.append(lev.Label)
                    lev.Height = levelHeight
                    lev.Placement.move(FreeCAD.Vector(0, 0, h))
                    if self.project and self.building:
                        from nativeifc import ifc_tools

                        lev = ifc_tools.aggregate(lev, self.building)
                    elif self.building:
                        self.building.addObject(lev)
                    h += levelHeight
                    for group in groups:
                        levGroup = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup")
                        levGroup.Label = group
                        if self.project:
                            from nativeifc import ifc_tools

                            ifc_tools.aggregate(levGroup, lev)
                        else:
                            lev.addObject(levGroup)
        FreeCAD.ActiveDocument.recompute()
        # fit zoom
        if outline:
            FreeCADGui.Selection.clearSelection()
            FreeCADGui.Selection.addSelection(outline)
            FreeCADGui.SendMsgToActiveView("ViewSelection")
            FreeCADGui.Selection.clearSelection()
        # aggregate layout group
        if self.building and grp:
            if hasattr(self.building, "IfcClass"):
                from nativeifc import ifc_tools

                ifc_tools.aggregate(grp, self.building)
        self.form.hide()
        FreeCAD.ActiveDocument.recompute()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.show()
        if self.form.radioNative3.isChecked():
            from nativeifc import ifc_status

            ifc_status.set_button(True, True)
        return self.reject()

    def addGroup(self):
        from PySide import QtCore, QtGui

        it = QtGui.QListWidgetItem(translate("BIM", "New Group"))
        it.setFlags(it.flags() | QtCore.Qt.ItemIsEditable)
        self.form.groupsList.addItem(it)

    def delGroup(self):
        r = self.form.groupsList.currentRow()
        if r >= 0:
            self.form.groupsList.takeItem(r)

    def savePreset(self):
        import Arch
        from PySide import QtGui

        res = QtGui.QInputDialog.getText(
            None,
            translate("BIM", "Save Preset"),
            translate("BIM", "Preset name"),
            QtGui.QLineEdit.Normal,
            "DefaultProject",
        )
        if res[1]:
            name = res[0]
            presetdir = os.path.join(FreeCAD.getUserAppDataDir(), "BIM")
            if not os.path.isdir(presetdir):
                os.mkdir(presetdir)

            groups = []
            for i in range(self.form.groupsList.count()):
                groups.append(self.form.groupsList.item(i).text())

            s = "# FreeCAD BIM Project setup preset " + name + "\n"
            s += "groupNewDocument=" + str(int(self.form.groupNewProject.isChecked())) + "\n"
            s += "projectName=" + self.form.projectName.text() + "\n"
            s += "groupSite=" + str(int(self.form.groupSite.isChecked())) + "\n"

            s += "siteName=" + self.form.siteName.text() + "\n"
            s += "siteAddress=" + self.form.siteAddress.text() + "\n"
            s += "siteLongitude=" + str(self.form.siteLongitude.value()) + "\n"
            s += "siteLatitude=" + str(self.form.siteLatitude.value()) + "\n"
            s += "siteDeviation=" + str(self.form.siteDeviation.value()) + "\n"
            s += "siteElevation=" + self.form.siteElevation.text() + "\n"

            s += "groupBuilding=" + str(int(self.form.groupBuilding.isChecked())) + "\n"
            s += "buildingName=" + self.form.buildingName.text() + "\n"
            s += "buildingUse=" + str(self.form.buildingUse.currentIndex()) + "\n"
            s += "buildingLength=" + self.form.buildingLength.text() + "\n"
            s += "buildingWidth=" + self.form.buildingWidth.text() + "\n"
            s += "countVAxes=" + str(self.form.countVAxes.value()) + "\n"
            s += "distVAxes=" + self.form.distVAxes.text() + "\n"
            s += "countHAxes=" + str(self.form.countHAxes.value()) + "\n"
            s += "distHAxes=" + self.form.distHAxes.text() + "\n"
            s += "countLevels=" + str(self.form.countLevels.value()) + "\n"
            s += "levelHeight=" + self.form.levelHeight.text() + "\n"
            s += "lineWidth=" + str(self.form.lineWidth.value()) + "\n"
            s += "lineColor=" + str(self.form.lineColor.property("color").getRgbF()[:3]) + "\n"
            s += "groups=" + ";;".join(groups) + "\n"
            s += "addHumanFigure=" + str(int(self.form.addHumanFigure.isChecked())) + "\n"

            f = open(os.path.join(presetdir, name + ".txt"), "w")
            f.write(s)
            f.close()
            self.fillPresets()

    def fillPresets(self):
        self.form.presets.clear()
        self.form.presets.addItem(translate("BIM", "User preset"))
        presetdir = os.path.join(FreeCAD.getUserAppDataDir(), "BIM")
        if os.path.isdir(presetdir):
            for f in os.listdir(presetdir):
                if f.endswith(".txt"):
                    self.form.presets.addItem(os.path.splitext(f)[0])

    def getPreset(self, preset):
        import Arch
        from PySide import QtGui

        preset = self.form.presets.itemText(preset)
        pfile = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", preset + ".txt")
        if os.path.exists(pfile):
            f = open(pfile, "r")
            buf = f.read()
            f.close()
            lines = buf.split("\n")
            for line in lines:
                if line:
                    if line[0] != "#":
                        s = line.split("=")
                        if s[0] == "groupNewDocument":
                            self.form.groupNewProject.setChecked(bool(int(s[1])))
                        elif s[0] == "projectName":
                            self.form.projectName.setText(s[1])
                        elif s[0] == "groupSite":
                            self.form.groupSite.setChecked(bool(int(s[1])))
                        elif s[0] == "siteName":
                            self.form.siteName.setText(s[1])
                        elif s[0] == "siteAddress":
                            self.form.siteAddress.setText(s[1])
                        elif s[0] == "siteLongitude":
                            self.form.siteLongitude.setValue(float(s[1]))
                        elif s[0] == "siteLatitude":
                            self.form.siteLatitude.setValue(float(s[1]))
                        elif s[0] == "siteDeviation":
                            self.form.siteDeviation.setValue(float(s[1]))
                        elif s[0] == "siteElevation":
                            self.form.siteElevation.setText(s[1])
                        elif s[0] == "groupBuilding":
                            self.form.groupBuilding.setChecked(bool(int(s[1])))
                        elif s[0] == "buildingName":
                            self.form.buildingName.setText(s[1])
                        elif s[0] == "buildingUse":
                            self.form.buildingUse.setCurrentIndex(int(s[1]))
                        elif s[0] == "buildingLength":
                            self.form.buildingLength.setText(s[1])
                        elif s[0] == "buildingWidth":
                            self.form.buildingWidth.setText(s[1])
                        elif s[0] == "countVAxes":
                            self.form.countVAxes.setValue(int(s[1]))
                        elif s[0] == "distVAxes":
                            self.form.distVAxes.setText(s[1])
                        elif s[0] == "countHAxes":
                            self.form.countHAxes.setValue(int(s[1]))
                        elif s[0] == "distHAxes":
                            self.form.distHAxes.setText(s[1])
                        elif s[0] == "countLevels":
                            self.form.countLevels.setValue(int(s[1]))
                        elif s[0] == "levelHeight":
                            self.form.levelHeight.setText(s[1])
                        elif s[0] == "lineWidth":
                            self.form.lineWidth.setValue(int(s[1]))
                        elif s[0] == "lineColor":
                            col = tuple([float(t) for t in s[1].strip("(").strip(")").split(",")])
                            col = QtGui.QColor.fromRgbF(*col)
                            self.form.lineColor.setProperty("color", col)
                        elif s[0] == "groups":
                            groups = s[1].split(";;")
                            self.form.groupsList.clear()
                            self.form.groupsList.addItems(groups)
                        elif s[0] == "addHumanFigure":
                            self.form.addHumanFigure.setChecked(bool(int(s[1])))

    def saveTemplate(self):
        """saves the contents of the current file as a template"""

        import WorkingPlane

        d = FreeCAD.ActiveDocument
        if not d:
            d = FreeCAD.newDocument()

        # build list of useful settings to store
        wp = WorkingPlane.get_working_plane()
        values = {}
        values["wpposition"] = str(wp.position)
        values["wpu"] = str(wp.u)
        values["wpv"] = str(wp.v)
        values["wpaxis"] = str(wp.axis)
        values["unit"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").GetInt("UserSchema", 0)
        )
        values["textsize"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetFloat(
                "textheight", 10
            )
        )
        values["textfont"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetString(
                "textfont", "Sans"
            )
        )
        values["dimstyle"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetInt("dimsymbol", 0)
        )
        values["arrowsize"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetFloat(
                "arrowsize", 5
            )
        )
        values["decimals"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").GetInt("Decimals", 2)
        )
        values["grid"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetFloat(
                "gridSpacing", 10
            )
        )
        values["squares"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetInt("gridEvery", 10)
        )
        values["linewidth"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").GetInt(
                "DefautShapeLineWidth", 2
            )
        )
        values["colFace"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").GetUnsigned(
                "DefaultShapeColor", 4294967295
            )
        )
        values["colLine"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").GetUnsigned(
                "DefaultShapeLineColor", 255
            )
        )
        values["colHelp"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Arch").GetUnsigned(
                "ColorHelpers", 674321151
            )
        )
        values["colConst"] = str(
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").GetUnsigned(
                "constructioncolor", 746455039
            )
        )

        d.Meta = values
        from PySide import QtGui

        filename = QtGui.QFileDialog.getSaveFileName(
            QtGui.QApplication.activeWindow(),
            translate("BIM", "Save template file"),
            None,
            "FreeCAD file (*.FCStd)",
        )
        if filename:
            filename = filename[0]
            if not filename.lower().endswith(".FCStd"):
                filename += ".FCStd"
            d.saveCopy(filename)
            FreeCAD.Console.PrintMessage(translate("BIM", "Template saved successfully") + "\n")
            self.reject()

    def loadTemplate(self):
        """loads the contents of a template into the current file"""

        from PySide import QtGui
        import FreeCADGui
        import WorkingPlane
        from FreeCAD import Vector  # required for following eval calls

        filename = QtGui.QFileDialog.getOpenFileName(
            QtGui.QApplication.activeWindow(),
            translate("BIM", "Open template file"),
            None,
            "FreeCAD file (*.FCStd)",
        )
        if filename:
            filename = filename[0]
            if FreeCAD.ActiveDocument:
                d = FreeCAD.ActiveDocument
                td = FreeCAD.openDocument(filename, True)  # hidden
                tname = td.Name
                values = td.Meta
                FreeCAD.closeDocument(tname)
                d.mergeProject(filename)
                FreeCADGui.ActiveDocument = FreeCADGui.getDocument(d.Name)  # fix b/c hidden doc
            else:
                d = FreeCAD.openDocument(filename)
                FreeCAD.ActiveDocument = d
                values = d.Meta
            bimunit = 0
            wp = WorkingPlane.get_working_plane()
            if "wpposition" in values:
                wp.position = eval(values["wpposition"])
            if "wpu" in values:
                wp.u = eval(values["wpu"])
            if "wpv" in values:
                wp.v = eval(values["wpv"])
            if "wpaxis" in values:
                wp.axis = eval(values["wpaxis"])
            wp._handle_custom(_hist_add=True)  # update the widget
            if "unit" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").SetInt(
                    "UserSchema", int(values["unit"])
                )
                if hasattr(FreeCAD.Units, "setSchema"):
                    FreeCAD.Units.setSchema(int(values["unit"]))
                    bimunit = [0, 2, 3, 3, 1, 5, 0, 4][int(values["unit"])]
            if "textsize" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetFloat(
                    "textheight", float(values["textsize"])
                )
                if hasattr(FreeCADGui, "draftToolBar"):
                    FreeCADGui.draftToolBar.fontsizeButton.setValue(float(values["textsize"]))
            if "textfont" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetString(
                    "textfont", values["textfont"]
                )
            if "dimstyle" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetInt(
                    "dimsymbol", int(values["dimstyle"])
                )
            if "arrowsize" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetFloat(
                    "arrowsize", float(values["arrowsize"])
                )
            if "decimals" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").SetInt(
                    "Decimals", int(values["decimals"])
                )
            if "grid" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetFloat(
                    "gridSpacing", float(values["grid"])
                )
            if "squares" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetInt(
                    "gridEvery", int(values["squares"])
                )
            if hasattr(FreeCADGui, "Snapper"):
                FreeCADGui.Snapper.setGrid()
            if "linewidth" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").SetInt(
                    "DefautShapeLineWidth", int(values["linewidth"])
                )
                if hasattr(FreeCADGui, "draftToolBar"):
                    FreeCADGui.draftToolBar.widthButton.setValue(int(values["linewidth"]))
            if "colFace" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").SetUnsigned(
                    "DefaultShapeColor", int(values["colFace"])
                )
            if "colLine" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").SetUnsigned(
                    "DefaultShapeLineColor", int(values["colLine"])
                )
            if "colHelp" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Arch").SetUnsigned(
                    "ColorHelpers", int(values["colHelp"])
                )
            if "colConst" in values:
                FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft").SetUnsigned(
                    "constructioncolor", int(values["colConst"])
                )

            # set the status bar widgets
            mw = FreeCADGui.getMainWindow()
            if mw:
                st = mw.statusBar()
                statuswidget = st.findChild(QtGui.QToolBar, "BIMStatusWidget")
                if statuswidget:
                    statuswidget.unitLabel.setText(statuswidget.unitsList[bimunit])
                    # change the unit of the nudge button
                    nudgeactions = statuswidget.nudge.menu().actions()
                    if bimunit in [2, 3, 5, 7]:
                        nudgelabels = statuswidget.nudgeLabelsI
                    else:
                        nudgelabels = statuswidget.nudgeLabelsM
                    for i in range(len(nudgelabels)):
                        nudgeactions[i].setText(nudgelabels[i])
                    if not "auto" in statuswidget.nudge.text().replace("&", "").lower():
                        statuswidget.nudge.setText(
                            FreeCAD.Units.Quantity(
                                statuswidget.nudge.text().replace("&", "")
                            ).UserString
                        )

            FreeCAD.Console.PrintMessage(
                translate("BIM", "Template successfully loaded into the current document") + "\n"
            )
            self.reject()


FreeCADGui.addCommand("BIM_ProjectManager", BIM_ProjectManager())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to trim and extend lines.

It also extends closed faces to create solids, that is, it can be used
to extrude a closed profile.

Make sure the snapping is active so that the extrusion is done following
the direction of a line, and up to the distance specified
by the snapping point.
"""
## @package gui_trimex
# \ingroup draftguitools
# \brief Provides GUI tools to trim and extend lines.

## \addtogroup draftguitools
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft
import Draft_rc
import DraftVecUtils
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftguitools.gui_trackers as trackers

from draftutils.messages import _msg, _err, _toolmsg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Trimex(gui_base_original.Modifier):
    """Gui Command for the Trimex tool.

    This tool trims or extends lines, wires and arcs,
    or extrudes single faces.

    SHIFT constrains to the last point
    or extrudes in direction to the face normal.
    """

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Trimex",
            "Accel": "T, R",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Trimex", "Trimex"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Trimex", "Trims or extends the selected object, or extrudes single faces"
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Trimex")
        self.edges = []
        self.placement = None
        self.ghost = []
        self.linetrack = None
        self.color = None
        self.width = None
        if self.ui:
            if not Gui.Selection.getSelection():
                self.ui.selectUi(on_close_call=self.finish)
                _msg(translate("draft", "Select objects to trim or extend"))
                self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)
            else:
                self.proceed()

    def proceed(self):
        """Proceed with execution of the command after proper selection."""
        if self.call:
            self.view.removeEventCallback("SoEvent", self.call)
        sel = Gui.Selection.getSelection()
        if len(sel) == 2:
            self.trimObjects(sel)
            self.finish()
            return
        self.obj = sel[0]
        self.ui.trimUi(title=translate("draft", self.featureName))
        self.linetrack = trackers.lineTracker()

        import DraftGeomUtils
        import Part

        if not hasattr(self.obj, "Shape"):
            self.obj = None
            self.finish()
            _err(translate("draft", "This object is not supported"))
            return
        if hasattr(self.obj, "Placement"):
            self.placement = self.obj.Placement
        if self.obj.Shape.Faces:
            sel = Gui.Selection.getSelectionEx("", 0)[0]
            self.obj = sel.Object
            if len(self.obj.Shape.Faces) == 1:
                # simple extrude mode, the object itself is extruded
                pass
            elif len(sel.SubObjects) == 1 and sel.SubObjects[0].ShapeType == "Face":
                # face extrude mode, a new object is created
                self.obj = self.doc.addObject("Part::Feature", "Face")
                self.obj.Shape = sel.SubObjects[0]
            else:
                self.obj = None
                self.finish()
                _err(translate("draft", "Only a single face can be extruded"))
                return
            self.extrudeMode = True
            self.normal = self.obj.Shape.Faces[0].normalAt(0.5, 0.5)
            self.ghost = [trackers.ghostTracker([self.obj]), trackers.lineTracker(dotted=True)]
            self.ghost += [trackers.lineTracker() for _ in self.obj.Shape.Vertexes]
        else:
            # normal wire trimex mode
            self.color = self.obj.ViewObject.LineColor
            self.width = self.obj.ViewObject.LineWidth
            if self.obj.Shape.Wires:
                self.edges = self.obj.Shape.Wires[0].Edges
                self.edges = Part.__sortEdges__(self.edges)
            else:
                self.edges = self.obj.Shape.Edges
            for e in self.edges:
                if isinstance(e.Curve, (Part.BSplineCurve, Part.BezierCurve)):
                    self.obj = None
                    self.finish()
                    _err(translate("draft", "Trimex does not support this object type"))
                    return
            self.obj.ViewObject.LineColor = (0.5, 0.5, 0.5)
            self.obj.ViewObject.LineWidth = 1
            self.extrudeMode = False
            self.ghost = []
            lc = self.color
            sc = (lc[0], lc[1], lc[2])
            sw = self.width
            for e in self.edges:
                if DraftGeomUtils.geomType(e) == "Line":
                    self.ghost.append(trackers.lineTracker(scolor=sc, swidth=sw))
                else:
                    self.ghost.append(trackers.arcTracker(scolor=sc, swidth=sw))
        if not self.ghost:
            self.obj = None
            self.finish()
            return
        for g in self.ghost:
            g.on()
        self.activePoint = 0
        self.nodes = []
        self.shift = False
        self.alt = False
        self.force = None
        self.cv = None
        self.call = self.view.addEventCallback("SoEvent", self.action)
        _toolmsg(translate("draft", "Pick distance"))

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif not self.ui.mouse:
            pass
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            self.shift = gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_constrain_key())
            self.alt = gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key())
            self.ctrl = gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_snap_key())
            if self.extrudeMode:
                arg["ShiftDown"] = False
            elif hasattr(Gui, "Snapper"):
                Gui.Snapper.setSelectMode(not self.ctrl)
            self.point, cp, info = gui_tool_utils.getPoint(self, arg)
            if gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_snap_key()):
                self.snapped = None
            else:
                self.snapped = self.view.getObjectInfo((arg["Position"][0], arg["Position"][1]))
            if self.extrudeMode:
                dist, ang = (self.extrude(self.shift), None)
            else:
                # If the geomType of the edge is "Line" ang will be None,
                # else dist will be None.
                dist, ang = self.redraw(self.point, self.snapped, self.shift, self.alt)

            if dist:
                self.ui.labelRadius.setText(translate("draft", "Distance"))
                self.ui.radiusValue.setToolTip(translate("draft", "Offset distance"))
                self.ui.setRadiusValue(dist, unit="Length")
            elif ang:
                self.ui.labelRadius.setText(translate("draft", "Angle"))
                self.ui.radiusValue.setToolTip(translate("draft", "Offset angle"))
                self.ui.setRadiusValue(ang, unit="Angle")
            else:
                # both dist and ang are None, this indicates an impossible
                # situation. Setting 0 with no unit will show "0 ??" and not
                # compute any value
                self.ui.setRadiusValue(0)
            self.ui.setFocus("radius")
            gui_tool_utils.redraw3DView()

        elif arg["Type"] == "SoMouseButtonEvent":
            if (arg["State"] == "DOWN") and (arg["Button"] == "BUTTON1"):
                cursor = arg["Position"]
                self.shift = gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_constrain_key())
                self.alt = gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key())
                if gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_snap_key()):
                    self.snapped = None
                else:
                    self.snapped = self.view.getObjectInfo((cursor[0], cursor[1]))
                self.trimObject()
                self.finish()

    def extrude(self, shift=False, real=False):
        """Redraw the ghost in extrude mode."""
        self.newpoint = self.obj.Shape.Faces[0].CenterOfMass
        dvec = self.point.sub(self.newpoint)
        if not shift:
            delta = DraftVecUtils.project(dvec, self.normal)
            if delta.Length < 1e-7:
                # Use the normal if self.newpoint is coplanar with the face:
                delta = self.normal * dvec.Length
        else:
            delta = dvec
        if self.force and delta.Length:
            ratio = self.force / delta.Length
            delta.multiply(ratio)
        if real:
            return delta
        self.ghost[0].trans.translation.setValue([delta.x, delta.y, delta.z])
        # Update the dotted lineTracker:
        self.ghost[1].p1(self.newpoint)
        self.ghost[1].p2(self.newpoint + dvec)
        # Update the vertex lineTrackers:
        for i in range(2, len(self.ghost)):
            base = self.obj.Shape.Vertexes[i - 2].Point
            self.ghost[i].p1(base)
            self.ghost[i].p2(base + delta)
        return delta.Length

    def redraw(self, point, snapped=None, shift=False, alt=False, real=None):
        """Redraw the ghost normally."""
        # initializing
        reverse = False
        for g in self.ghost:
            g.off()
        if real:
            newedges = []

        import DraftGeomUtils
        import Part

        # finding the active point
        vlist = []
        for e in self.edges:
            vlist.append(e.Vertexes[0].Point)
        vlist.append(self.edges[-1].Vertexes[-1].Point)
        if shift:
            npoint = self.activePoint
        else:
            npoint = DraftGeomUtils.findClosest(point, vlist)
        if npoint > len(self.edges) / 2:
            reverse = True
        if alt:
            reverse = not reverse
        self.activePoint = npoint

        # sorting out directions
        if reverse and (npoint > 0):
            npoint = npoint - 1
        if npoint > len(self.edges) - 1:
            edge = self.edges[-1]
            ghost = self.ghost[-1]
        else:
            edge = self.edges[npoint]
            ghost = self.ghost[npoint]
        if reverse:
            v1 = edge.Vertexes[-1].Point
            v2 = edge.Vertexes[0].Point
        else:
            v1 = edge.Vertexes[0].Point
            v2 = edge.Vertexes[-1].Point

        # snapping
        if snapped:
            parent = snapped.get("ParentObject", None)
            if parent:
                subname = snapped["SubName"]
            else:
                parent = self.doc.getObject(snapped["Object"])
                subname = snapped["Component"]
            shape = Part.getShape(parent, subname, needSubElement=True, noElementMap=True)
            if shape.Edges:
                pts = []
                for e in shape.Edges:
                    int = DraftGeomUtils.findIntersection(edge, e, True, True)
                    if int:
                        pts.extend(int)
                if pts:
                    point = pts[DraftGeomUtils.findClosest(point, pts)]

        # modifying active edge
        if DraftGeomUtils.geomType(edge) == "Line":
            ang = None
            ve = DraftGeomUtils.vec(edge)
            chord = v1.sub(point)
            n = ve.cross(chord)
            if n.Length == 0:
                self.newpoint = point
            else:
                perp = ve.cross(n)
                proj = DraftVecUtils.project(chord, perp)
                self.newpoint = App.Vector.add(point, proj)
            dist = v1.sub(self.newpoint).Length
            ghost.p1(self.newpoint)
            ghost.p2(v2)
            if real:
                if self.force:
                    ray = self.newpoint.sub(v1)
                    if ray.Length:
                        ray.multiply(self.force / ray.Length)
                    self.newpoint = App.Vector.add(v1, ray)
                newedges.append(Part.LineSegment(self.newpoint, v2).toShape())
        else:
            dist = None
            center = edge.Curve.Center
            rad = edge.Curve.Radius
            ang1 = DraftVecUtils.angle(v2.sub(center))
            ang2 = DraftVecUtils.angle(point.sub(center))
            _rot_rad = DraftVecUtils.rotate(App.Vector(rad, 0, 0), -ang2)
            self.newpoint = App.Vector.add(center, _rot_rad)
            ang = math.degrees(-ang2)
            # if ang1 > ang2:
            #     ang1, ang2 = ang2, ang1
            # print("last calculated:",
            #       math.degrees(-ang1),
            #       math.degrees(-ang2))
            ghost.setEndAngle(-ang2)
            ghost.setStartAngle(-ang1)
            ghost.setCenter(center)
            ghost.setRadius(rad)
            if real:
                if self.force:
                    angle = math.radians(self.force)
                    newray = DraftVecUtils.rotate(App.Vector(rad, 0, 0), -angle)
                    self.newpoint = App.Vector.add(center, newray)
                chord = self.newpoint.sub(v2)
                perp = chord.cross(App.Vector(0, 0, 1))
                scaledperp = DraftVecUtils.scaleTo(perp, rad)
                midpoint = App.Vector.add(center, scaledperp)
                _sh = Part.Arc(self.newpoint, midpoint, v2).toShape()
                newedges.append(_sh)
        ghost.on()

        # resetting the edges
        if not reverse:
            li = list(range(npoint + 1, len(self.edges)))
        else:
            li = list(range(npoint - 1, -1, -1))
        for i in li:
            edge = self.edges[i]
            ghost = self.ghost[i]
            if DraftGeomUtils.geomType(edge) == "Line":
                ghost.p1(edge.Vertexes[0].Point)
                ghost.p2(edge.Vertexes[-1].Point)
            else:
                ang1 = DraftVecUtils.angle(edge.Vertexes[0].Point.sub(center))
                ang2 = DraftVecUtils.angle(edge.Vertexes[-1].Point.sub(center))
                # if ang1 > ang2:
                #     ang1, ang2 = ang2, ang1
                ghost.setEndAngle(-ang2)
                ghost.setStartAngle(-ang1)
                ghost.setCenter(edge.Curve.Center)
                ghost.setRadius(edge.Curve.Radius)
            if real:
                newedges.append(edge)
            ghost.on()

        # finishing
        if real:
            return newedges
        else:
            return [dist, ang]

    def trimObject(self):
        """Trim the actual object."""
        import Part

        if self.extrudeMode:
            delta = self.extrude(self.shift, real=True)
            # print("delta", delta)
            self.doc.openTransaction("Extrude")
            Gui.addModule("Draft")
            obj = Draft.extrude(self.obj, delta, solid=True)
            self.doc.commitTransaction()
            self.obj = obj
        else:
            edges = self.redraw(self.point, self.snapped, self.shift, self.alt, real=True)
            newshape = Part.Wire(edges)
            self.doc.openTransaction("Trim/extend")
            if utils.getType(self.obj) in ["Wire", "BSpline"]:
                p = []
                if self.placement:
                    invpl = self.placement.inverse()
                for v in newshape.Vertexes:
                    np = v.Point
                    if self.placement:
                        np = invpl.multVec(np)
                    p.append(np)
                self.obj.Points = p
            elif utils.getType(self.obj) == "Part::Line":
                p = []
                if self.placement:
                    invpl = self.placement.inverse()
                for v in newshape.Vertexes:
                    np = v.Point
                    if self.placement:
                        np = invpl.multVec(np)
                    p.append(np)
                if (p[0].x == self.obj.X1) and (p[0].y == self.obj.Y1) and (p[0].z == self.obj.Z1):
                    self.obj.X2 = p[-1].x
                    self.obj.Y2 = p[-1].y
                    self.obj.Z2 = p[-1].z
                elif (
                    (p[-1].x == self.obj.X1)
                    and (p[-1].y == self.obj.Y1)
                    and (p[-1].z == self.obj.Z1)
                ):
                    self.obj.X2 = p[0].x
                    self.obj.Y2 = p[0].y
                    self.obj.Z2 = p[0].z
                elif (
                    (p[0].x == self.obj.X2) and (p[0].y == self.obj.Y2) and (p[0].z == self.obj.Z2)
                ):
                    self.obj.X1 = p[-1].x
                    self.obj.Y1 = p[-1].y
                    self.obj.Z1 = p[-1].z
                else:
                    self.obj.X1 = p[0].x
                    self.obj.Y1 = p[0].y
                    self.obj.Z1 = p[0].z
            elif utils.getType(self.obj) == "Circle":
                angles = self.ghost[0].getAngles()
                # print("original", self.obj.FirstAngle," ",self.obj.LastAngle)
                # print("new", angles)
                if angles[0] > angles[1]:
                    angles = (angles[1], angles[0])
                self.obj.FirstAngle = angles[0]
                self.obj.LastAngle = angles[1]
            else:
                self.obj.Shape = newshape
            self.doc.commitTransaction()
        self.doc.recompute()
        for g in self.ghost:
            g.off()

    def trimObjects(self, objectslist):
        """Attempt to trim two objects together."""
        import Part
        import DraftGeomUtils

        wires = []
        for obj in objectslist:
            if not utils.getType(obj) in ["Wire", "Circle"]:
                _err(
                    translate(
                        "draft",
                        "Unable to trim these objects, " "only Draft wires and arcs are supported",
                    )
                )
                return
            if len(obj.Shape.Wires) > 1:
                _err(translate("draft", "Unable to trim these objects, " "too many wires"))
                return
            if len(obj.Shape.Wires) == 1:
                wires.append(obj.Shape.Wires[0])
            else:
                wires.append(Part.Wire(obj.Shape.Edges))
        ints = []
        edge1 = None
        edge2 = None
        for i1, e1 in enumerate(wires[0].Edges):
            for i2, e2 in enumerate(wires[1].Edges):
                i = DraftGeomUtils.findIntersection(e1, e2, dts=False)
                if len(i) == 1:
                    ints.append(i[0])
                    edge1 = i1
                    edge2 = i2
        if not ints:
            _err(translate("draft", "These objects do not intersect"))
            return
        if len(ints) != 1:
            _err(translate("draft", "Too many intersection points"))
            return

        v11 = wires[0].Vertexes[0].Point
        v12 = wires[0].Vertexes[-1].Point
        v21 = wires[1].Vertexes[0].Point
        v22 = wires[1].Vertexes[-1].Point
        if DraftVecUtils.closest(ints[0], [v11, v12]) == 1:
            last1 = True
        else:
            last1 = False
        if DraftVecUtils.closest(ints[0], [v21, v22]) == 1:
            last2 = True
        else:
            last2 = False
        for i, obj in enumerate(objectslist):
            if i == 0:
                ed = edge1
                la = last1
            else:
                ed = edge2
                la = last2
            if utils.getType(obj) == "Wire":
                if la:
                    pts = obj.Points[: ed + 1] + ints
                else:
                    pts = ints + obj.Points[ed + 1 :]
                obj.Points = pts
            else:
                vec = ints[0].sub(obj.Placement.Base)
                vec = obj.Placement.inverse().Rotation.multVec(vec)
                _x = App.Vector(1, 0, 0)
                _ang = -DraftVecUtils.angle(
                    vec, obj.Placement.Rotation.multVec(_x), obj.Shape.Edges[0].Curve.Axis
                )
                ang = math.degrees(_ang)
                if la:
                    obj.LastAngle = ang
                else:
                    obj.FirstAngle = ang
        self.doc.recompute()

    def finish(self, cont=False):
        """Terminate the operation of the Trimex tool."""
        self.end_callbacks(self.call)
        self.force = None
        if self.ui:
            if self.linetrack:
                self.linetrack.finalize()
            if self.ghost:
                for g in self.ghost:
                    g.finalize()
            if self.obj:
                if self.color:
                    self.obj.ViewObject.LineColor = self.color
                if self.width:
                    self.obj.ViewObject.LineWidth = self.width
                gui_utils.select(self.obj)
        super().finish()

    def numericRadius(self, dist):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.force = dist
        self.trimObject()
        self.finish()


Gui.addCommand("Draft_Trimex", Trimex())

## @}



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint electrostatic potential document object"
__author__ = "Markus Hovorka, Bernd Hahnebach, Uwe Stöhr, Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package constraint_electrostaticpotential
#  \ingroup FEM
#  \brief constraint electrostatic potential object

from FreeCAD import Base

from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class ConstraintElectrostaticPotential(base_fempythonobject.BaseFemPythonObject):

    Type = "Fem::ConstraintElectrostaticPotential"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyElectricPotential",
                name="Potential",
                group="Parameter",
                doc="Electric Potential",
                value="1 V",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="PotentialEnabled",
                group="Parameter",
                doc="Enable electric potential",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectricPotential",
                name="AV_re",
                group="Electromagnetic Potential",
                doc="Real part of scalar potential",
                value="0 V",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectricPotential",
                name="AV_im",
                group="Electromagnetic Potential",
                doc="Imaginary part of scalar potential",
                value="0 V",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectromagneticPotential",
                name="AV_re_1",
                group="Electromagnetic Potential",
                doc="Real part of vector potential x-component",
                value="0 Wb/m",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectromagneticPotential",
                name="AV_re_2",
                group="Electromagnetic Potential",
                doc="Real part of vector potential y-component",
                value="0 Wb/m",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectromagneticPotential",
                name="AV_re_3",
                group="Electromagnetic Potential",
                doc="Real part of vector potential z-component",
                value="0 Wb/m",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectromagneticPotential",
                name="AV_im_1",
                group="Electromagnetic Potential",
                doc="Imaginary part of vector potential x-component",
                value="0 Wb/m",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectromagneticPotential",
                name="AV_im_2",
                group="Electromagnetic Potential",
                doc="Imaginary part of vector potential y-component",
                value="0 Wb/m",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyElectromagneticPotential",
                name="AV_im_3",
                group="Electromagnetic Potential",
                doc="Imaginary part of vector potential z-component",
                value="0 Wb/m",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertySurfaceChargeDensity",
                name="ElectricFluxDensity",
                group="Parameter",
                doc="Electric displacement field D.\n"
                + "For interfaces, it represents the difference\n"
                + "between the normal component in the two media",
                value="0 C/m^2",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="BoundaryCondition",
                group="Parameter",
                doc="Set boundary condition type",
                value=["Dirichlet", "Neumann"],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableAV",
                group="Electromagnetic Potential",
                doc="Enable scalar potential boundary condition",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableAV_1",
                group="Electromagnetic Potential",
                doc="Enable vector potential x-component boundary condition",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableAV_2",
                group="Electromagnetic Potential",
                doc="Enable vector potential y-component boundary condition",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="EnableAV_3",
                group="Electromagnetic Potential",
                doc="Enable vector potential z-component boundary condition",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="PotentialConstant",
                group="Parameter",
                doc="",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="ElectricInfinity",
                group="Parameter",
                doc="Electric Infinity",
                value=False,
            )
        )

        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="ElectricForcecalculation",
                group="Parameter",
                doc="Electric force calculation",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyInteger",
                name="CapacitanceBody",
                group="Parameter",
                doc="Capacitance body",
                value=1,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CapacitanceBodyEnabled",
                group="Parameter",
                doc="Capacitance body enabled",
                value=False,
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

            # convert old potential float to Volt
            if prop.name == "Potential":
                prop.handle_change_type(
                    obj,
                    old_type="App::PropertyFloat",
                    convert_old_value=lambda x: "{} V".format(1e6 * x),
                )
            # fix unit for magnetic vector potential properties
            if prop.name in ("AV_re_1", "AV_re_2", "AV_re_3", "AV_im_1", "AV_im_2", "AV_im_3"):
                prop.handle_change_type(
                    obj,
                    old_type="App::PropertyElectricPotential",
                    convert_old_value=lambda x: "{} Wb/m".format(x.getValueAs("V").Value),
                )

        # enable electromagnetic properties from old properties
        try:
            obj.EnableAV_1 = not obj.getPropertyByName(
                "AV_re_1_Disabled"
            ) or not obj.getPropertyByName("AV_im_1_Disabled")
            obj.EnableAV_2 = not obj.getPropertyByName(
                "AV_re_2_Disabled"
            ) or not obj.getPropertyByName("AV_im_2_Disabled")
            obj.EnableAV_3 = not obj.getPropertyByName(
                "AV_re_3_Disabled"
            ) or not obj.getPropertyByName("AV_im_3_Disabled")
            obj.EnableAV = not obj.getPropertyByName("AV_im_Disabled")

            # remove old properties
            obj.setPropertyStatus("AV_re_1_Disabled", "-LockDynamic")
            obj.removeProperty("AV_re_1_Disabled")
            obj.setPropertyStatus("AV_re_2_Disabled", "-LockDynamic")
            obj.removeProperty("AV_re_2_Disabled")
            obj.setPropertyStatus("AV_re_3_Disabled", "-LockDynamic")
            obj.removeProperty("AV_re_3_Disabled")
            obj.setPropertyStatus("AV_im_1_Disabled", "-LockDynamic")
            obj.removeProperty("AV_im_1_Disabled")
            obj.setPropertyStatus("AV_im_2_Disabled", "-LockDynamic")
            obj.removeProperty("AV_im_2_Disabled")
            obj.setPropertyStatus("AV_im_3_Disabled", "-LockDynamic")
            obj.removeProperty("AV_im_3_Disabled")
            obj.setPropertyStatus("AV_im_Disabled", "-LockDynamic")
            obj.removeProperty("AV_im_Disabled")

        except Base.PropertyError:
            pass

        # set electric flux density from old surface charge density
        try:
            obj.ElectricFluxDensity = obj.getPropertyByName("SurfaceChargeDensity")
            obj.setPropertyStatus("SurfaceChargeDensity", "-LockDynamic")
            obj.removeProperty("SurfaceChargeDensity")

        except Base.PropertyError:
            pass



#!/usr/bin/env python3


import FreeCAD
import os
import unittest


class DrawHatchTest(unittest.TestCase):
    def setUp(self):
        """Creates a page and view"""
        self.path = os.path.dirname(os.path.abspath(__file__))
        print("TDHatch path: " + self.path)
        templateFileSpec = self.path + "/TestTemplate.svg"

        FreeCAD.newDocument("TDHatch")
        FreeCAD.setActiveDocument("TDHatch")
        FreeCAD.ActiveDocument = FreeCAD.getDocument("TDHatch")

        # make source feature
        box = FreeCAD.ActiveDocument.addObject("Part::Box", "Box")

        # make a page
        self.page = FreeCAD.ActiveDocument.addObject("TechDraw::DrawPage", "Page")
        FreeCAD.ActiveDocument.addObject("TechDraw::DrawSVGTemplate", "Template")
        FreeCAD.ActiveDocument.Template.Template = templateFileSpec
        FreeCAD.ActiveDocument.Page.Template = FreeCAD.ActiveDocument.Template
        self.page.Scale = 5.0
        # page.ViewObject.show()  #unit tests run in console mode

        # make Views
        self.view = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewPart", "View")
        FreeCAD.ActiveDocument.View.Source = [box]
        self.page.addView(self.view)
        FreeCAD.ActiveDocument.recompute()

    def tearDown(self):
        FreeCAD.closeDocument("TDHatch")

    def testMakeHatchCase(self):
        """Tests if hatch area can be added to view"""
        # make hatch
        print("making hatch")
        hatch = FreeCAD.ActiveDocument.addObject("TechDraw::DrawHatch", "Hatch")
        hatch.Source = (self.view, ["Face0"])
        hatchFileSpec = self.path + "/TestHatch.svg"
        # comment out to use default from preferences
        hatch.HatchPattern = (
            hatchFileSpec
        )
        print("finished hatch")
        FreeCAD.ActiveDocument.recompute()

        self.assertTrue("Up-to-date" in hatch.State)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import FreeCAD
import FreeCADGui

translate = FreeCAD.Qt.translate


def preferences():
    return FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Assembly")


class PreferencesPage:
    def __init__(self, parent=None):
        self.form = FreeCADGui.PySideUic.loadUi(":preferences/Assembly.ui")

    def saveSettings(self):
        pref = preferences()
        pref.SetBool("LeaveEditWithEscape", self.form.checkBoxEnableEscape.isChecked())
        pref.SetBool("LogSolverDebug", self.form.checkBoxSolverDebug.isChecked())
        pref.SetInt("GroundFirstPart", self.form.groundFirstPart.currentIndex())

    def loadSettings(self):
        pref = preferences()
        self.form.checkBoxEnableEscape.setChecked(pref.GetBool("LeaveEditWithEscape", True))
        self.form.checkBoxSolverDebug.setChecked(pref.GetBool("LogSolverDebug", False))
        self.form.groundFirstPart.clear()
        self.form.groundFirstPart.addItem(translate("Assembly", "Ask"))
        self.form.groundFirstPart.addItem(translate("Assembly", "Always"))
        self.form.groundFirstPart.addItem(translate("Assembly", "Never"))
        self.form.groundFirstPart.setCurrentIndex(pref.GetInt("GroundFirstPart", 0))



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Contains the Addon Manager's preferences dialog management class"""

import os
from enum import IntEnum
import ipaddress
import re
from typing import Tuple

import addonmanager_freecad_interface as fci
from addonmanager_preferences_migrations import migrate_proxy_settings_2025
from NetworkManager import ForceReinitializeNetworkManager

from PySideWrapper import QtCore, QtGui, QtWidgets, QtNetwork, QtSvg

translate = fci.translate

# pylint: disable=too-few-public-methods


def is_ip(host: str) -> bool:
    try:
        ipaddress.ip_address(host)
        return True
    except ValueError:
        return False


# Simple regex for valid domain names. Not exhaustive, but hopefully good enough.
# * Each label starts/ends with alphanumeric, can contain hyphens.
# * TLD must be at least 2 letters.
DOMAIN_REGEX = re.compile(r"^(?:[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?\.)+[A-Za-z]{2,}$")


def is_domain(host: str) -> bool:
    return DOMAIN_REGEX.match(host) is not None


def is_valid_host(host: str) -> bool:
    if not host:
        return False
    return is_ip(host) or is_domain(host)


def test_proxy_connection(
    proxy: QtNetwork.QNetworkProxy | QtNetwork.QNetworkProxy.ProxyType,
) -> Tuple[bool, str]:

    nam = QtNetwork.QNetworkAccessManager()
    nam.setProxy(proxy)
    url = fci.Preferences().get("status_test_url")
    req = QtNetwork.QNetworkRequest(QtCore.QUrl(url))
    req.setAttribute(QtNetwork.QNetworkRequest.Http2AllowedAttribute, False)
    reply = nam.get(req)
    loop = QtCore.QEventLoop()
    timer = QtCore.QTimer()
    timer.setSingleShot(True)
    timer.timeout.connect(loop.quit)
    reply.finished.connect(loop.quit)
    timer.start(3000)
    if hasattr(loop, "exec"):
        loop.exec()
    else:
        loop.exec_()  # Qt5
    if timer.isActive():
        timer.stop()
    else:
        reply.abort()
        reply.deleteLater()
        nam.deleteLater()
        return False, translate("AddonsInstaller", "Proxy test timed out: no connection made.")

    if reply.error() != QtNetwork.QNetworkReply.NoError:
        msg = f"QtNetwork error: {reply.error()} - {reply.errorString()}"
        reply.deleteLater()
        nam.deleteLater()
        return (
            False,
            translate("AddonsInstaller", "Proxy test returned an error: no connection made.\n")
            + msg,
        )

    status = reply.attribute(QtNetwork.QNetworkRequest.HttpStatusCodeAttribute)
    status = int(status) if status is not None else None
    reason = reply.attribute(QtNetwork.QNetworkRequest.HttpReasonPhraseAttribute)
    reason = (
        bytes(reason).decode("utf-8") if isinstance(reason, QtCore.QByteArray) else (reason or "")
    )

    reply.deleteLater()
    nam.deleteLater()

    # Success criteria: 2xx or 3xx typically indicates the proxy path worked.
    if status is not None and 200 <= status < 400:
        return True, translate("AddonsInstaller", "Proxy test succeeded, connection established.")
    if status == 407:
        return False, translate(
            "AddonsInstaller",
            "Proxy requires authentication. The Addon Manager does not support this.",
        )
    return False, translate("AddonsInstaller", "Proxy connection failed with code {}: {}.").format(
        status, reason
    )


class AddonManagerOptions:
    """A class containing a form element that is inserted as a FreeCAD preference page."""

    class ProxyType(IntEnum):
        """This is an IntEnum to continue to support Python 3.8 (e.g., FreeCAD 0.21) but can be
        converted to a StrEnum when we can raise minimum support to Python 3.11."""

        none = 0
        system = 1
        custom = 2

        def __str__(self):
            if self == self.none:
                return "none"
            if self == self.system:
                return "system"
            if self == self.custom:
                return "custom"
            return "unknown"

        @staticmethod
        def from_string(s: str) -> "AddonManagerOptions.ProxyType":
            if s.lower() == "none":
                return AddonManagerOptions.ProxyType.none
            if s.lower() == "system":
                return AddonManagerOptions.ProxyType.system
            if s.lower() == "custom":
                return AddonManagerOptions.ProxyType.custom
            return AddonManagerOptions.ProxyType.none

    class ProxyTestStatus(IntEnum):
        untested = 0
        testing = 1
        success = 2
        failure = 3

    def __init__(self, _=None):
        self.form = fci.loadUi(os.path.join(os.path.dirname(__file__), "AddonManagerOptions.ui"))
        self.form.setObjectName("AddonManager_PreferencesTab")
        self.table_model = CustomRepoDataModel()
        self.form.customRepositoriesTableView.setModel(self.table_model)
        icon_path = os.path.join(os.path.dirname(__file__), "Resources", "icons")
        self.form.addCustomRepositoryButton.setIcon(
            QtGui.QIcon.fromTheme("add", QtGui.QIcon(os.path.join(icon_path, "list-add.svg")))
        )
        self.form.removeCustomRepositoryButton.setIcon(
            QtGui.QIcon.fromTheme("remove", QtGui.QIcon(os.path.join(icon_path, "list-remove.svg")))
        )

        self.form.customRepositoriesTableView.horizontalHeader().setStretchLastSection(False)
        self.form.customRepositoriesTableView.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.Stretch
        )
        self.form.customRepositoriesTableView.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.ResizeToContents
        )
        line_height = self.form.customRepositoriesTableView.verticalHeader().defaultSectionSize()
        self.form.customRepositoriesTableView.setFixedHeight(line_height * 6.5)

        self.form.addCustomRepositoryButton.clicked.connect(self._add_custom_repo_clicked)
        self.form.removeCustomRepositoryButton.clicked.connect(self._remove_custom_repo_clicked)
        self.form.customRepositoriesTableView.doubleClicked.connect(self._row_double_clicked)

        self.form.proxyGroupBox.toggled.connect(self._proxy_state_changed)
        self.form.systemProxyButton.clicked.connect(self._proxy_type_changed)
        self.form.customProxyButton.clicked.connect(self._proxy_type_changed)
        self.form.proxyTestButton.clicked.connect(self._test_proxy)
        self.form.proxyTestButton.setIcon(
            QtGui.QIcon.fromTheme(
                "view-refresh", QtGui.QIcon(os.path.join(icon_path, "view-refresh"))
            )
        )
        self.form.proxyHostLineEdit.textChanged.connect(self._proxy_changed)
        self.form.proxyPortLineEdit.textChanged.connect(self._proxy_changed)
        self._set_proxy_test_button_state(AddonManagerOptions.ProxyTestStatus.untested)

        int_validator = QtGui.QIntValidator(1, 65535)  # Valid range for port numbers
        self.form.proxyPortLineEdit.setValidator(int_validator)

        hostname_regex = QtCore.QRegularExpression(
            r"^[A-Za-z0-9]+(?:[-A-Za-z0-9]*[A-Za-z0-9])?(?:\.[A-Za-z0-9]+(?:[-A-Za-z0-9]*[A-Za-z0-9])?)*$"
        )
        self.form.proxyHostLineEdit.setValidator(QtGui.QRegularExpressionValidator(hostname_regex))

        fm = QtGui.QFontMetrics(self.form.proxyPortLineEdit.font())
        char_width = fm.horizontalAdvance("M")
        target_width = char_width * 5 + 10  # Five chars max, plus some padding
        self.form.proxyPortLineEdit.setFixedWidth(target_width)

        renderer = QtSvg.QSvgRenderer(os.path.join(icon_path, "regex_bad.svg"))
        pixmap = QtGui.QPixmap(char_width, char_width)
        pixmap.fill(QtCore.Qt.transparent)  # keep background transparent
        painter = QtGui.QPainter(pixmap)
        renderer.render(painter)  # renders the whole SVG scaled to pixmap
        painter.end()

        self.form.proxyHostInvalidIcon.setPixmap(pixmap)
        self.form.proxyHostInvalidIcon.setToolTip(translate("AddonsInstaller", "Invalid hostname"))
        self.form.proxyHostInvalidIcon.hide()

        self.form.proxyStatusTestGroupBox.setVisible(False)

        migrate_proxy_settings_2025()

    def reconfigure_proxy_ui(self, proxy_type: ProxyType):
        if proxy_type == AddonManagerOptions.ProxyType.none:
            self.form.proxyGroupBox.setChecked(False)
            self.form.proxyHostLineEdit.setPlaceholderText(translate("AddonsInstaller", "No proxy"))
            self.form.proxyPortLineEdit.setPlaceholderText(translate("AddonsInstaller", "n/a"))
        else:
            self.form.proxyGroupBox.setChecked(True)
            self.form.proxyHostLineEdit.setPlaceholderText(
                translate("AddonsInstaller", "proxy.example.com")
            )
            self.form.proxyPortLineEdit.setPlaceholderText("8080")
            if proxy_type == AddonManagerOptions.ProxyType.system:
                self.form.systemProxyButton.setChecked(True)
                self.form.proxyHostLineEdit.setEnabled(False)
                self.form.proxyPortLineEdit.setEnabled(False)
            else:
                self.form.customProxyButton.setChecked(True)
                self.form.proxyHostLineEdit.setEnabled(True)
                self.form.proxyPortLineEdit.setEnabled(True)

    def fill_proxy_host_settings_from_preferences(self):
        proxy_type = fci.Preferences().get("proxy_type")
        if proxy_type == AddonManagerOptions.ProxyType.none:
            self.form.proxyHostLineEdit.setText(translate("AddonsInstaller", "No proxy"))
            self.form.proxyPortLineEdit.setText("8080")
        elif proxy_type == AddonManagerOptions.ProxyType.system:
            self.fill_proxy_with_system_settings()
        else:
            self.form.proxyHostLineEdit.setText(fci.Preferences().get("proxy_host"))
            self.form.proxyPortLineEdit.setText(str(fci.Preferences().get("proxy_port")))

    def _custom_activated(self):
        self.form.proxyHostLineEdit.setText(fci.Preferences().get("proxy_host"))
        self.form.proxyPortLineEdit.setText(str(fci.Preferences().get("proxy_port")))

    def fill_proxy_with_system_settings(self):
        url = fci.Preferences().get("status_test_url")
        query = QtNetwork.QNetworkProxyQuery(QtCore.QUrl(url))
        proxy = QtNetwork.QNetworkProxyFactory.systemProxyForQuery(query)
        if proxy and proxy[0] and proxy[0].hostName() and proxy[0].port() > 0:
            self.form.proxyHostLineEdit.setText(proxy[0].hostName())
            self.form.proxyPortLineEdit.setText(str(proxy[0].port()))
        else:
            self.form.proxyHostLineEdit.setText(translate("AddonsInstaller", "System has no proxy"))
            self.form.proxyPortLineEdit.setText("8080")

    def _proxy_type_changed(self):
        """Callback: when the proxy type is changed, update the UI accordingly"""
        proxy_type = self._proxy_type_from_ui()
        self.reconfigure_proxy_ui(proxy_type)
        self.fill_proxy_host_settings_from_preferences()
        self._proxy_changed()
        if proxy_type == AddonManagerOptions.ProxyType.custom:
            self._custom_activated()

    def _proxy_state_changed(self):
        """Callback: when the proxy state is changed, update the UI accordingly"""
        proxy_type = self._proxy_type_from_ui()
        self.reconfigure_proxy_ui(proxy_type)
        self._proxy_changed()
        if proxy_type == AddonManagerOptions.ProxyType.none:
            self.form.proxyHostLineEdit.setText(translate("AddonsInstaller", "No proxy"))
        elif proxy_type == AddonManagerOptions.ProxyType.system:
            self.fill_proxy_with_system_settings()
        else:
            self.fill_proxy_host_settings_from_preferences()

    def _proxy_changed(self):
        self._set_proxy_test_button_state(AddonManagerOptions.ProxyTestStatus.untested)
        self.form.proxyStatusTestGroupBox.setVisible(False)
        if self._proxy_type_from_ui() == AddonManagerOptions.ProxyType.custom:
            if is_valid_host(self.form.proxyHostLineEdit.text()):
                self.form.proxyTestButton.setEnabled(True)
                self.form.proxyHostInvalidIcon.hide()
            else:
                self.form.proxyTestButton.setEnabled(False)
                self.form.proxyHostInvalidIcon.show()
        else:
            self.form.proxyHostInvalidIcon.hide()

    def _test_proxy(self):
        """Callback: when the test proxy button is clicked, test the proxy settings"""
        self._set_proxy_test_button_state(AddonManagerOptions.ProxyTestStatus.testing)
        self.form.proxyStatusTestGroupBox.setVisible(True)
        self.form.proxyStatusTestOutputLabel.setText(
            translate("AddonsInstaller", "Testing proxy connection…")
        )
        proxy_type = self._proxy_type_from_ui()
        if proxy_type == AddonManagerOptions.ProxyType.none:
            proxy = QtNetwork.QNetworkProxy.NoProxy
        else:
            if proxy_type == AddonManagerOptions.ProxyType.system:
                url = fci.Preferences().get("status_test_url")
                query = QtNetwork.QNetworkProxyQuery(QtCore.QUrl(url))
                proxies = QtNetwork.QNetworkProxyFactory.systemProxyForQuery(query)
                if proxies and proxies[0] and proxies[0].hostName() and proxies[0].port() > 0:
                    proxy = proxies[0]
                else:
                    proxy = QtNetwork.QNetworkProxy.NoProxy
            else:
                scheme = QtNetwork.QNetworkProxy.HttpProxy
                host = self.form.proxyHostLineEdit.text()
                port = int(self.form.proxyPortLineEdit.text())
                proxy = QtNetwork.QNetworkProxy(scheme, host, port)

        status, message = test_proxy_connection(proxy)
        self.form.proxyStatusTestOutputLabel.setText(message)
        self._set_proxy_test_button_state(
            AddonManagerOptions.ProxyTestStatus.success
            if status
            else AddonManagerOptions.ProxyTestStatus.failure
        )

    def _set_proxy_test_button_state(self, state: ProxyTestStatus):
        icon_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "Resources", "icons")
        icon = QtGui.QIcon.fromTheme(
            "view-refresh", QtGui.QIcon(os.path.join(icon_path, "view-refresh"))
        )
        if state == AddonManagerOptions.ProxyTestStatus.success:
            icon = QtGui.QIcon.fromTheme("ok", QtGui.QIcon(os.path.join(icon_path, "regex_ok.svg")))
        elif state == AddonManagerOptions.ProxyTestStatus.failure:
            icon = QtGui.QIcon.fromTheme(
                "cancel", QtGui.QIcon(os.path.join(icon_path, "regex_bad.svg"))
            )
        if state == AddonManagerOptions.ProxyTestStatus.testing:
            self.form.proxyTestButton.setEnabled(False)
        else:
            self.form.proxyTestButton.setEnabled(True)
        self.form.proxyTestButton.setIcon(icon)

    def saveSettings(self):
        """Required function: called by the preferences dialog when Apply or Save is clicked,
        saves out the preference data by reading it from the widgets."""
        for widget in self.form.children():
            self.recursive_widget_saver(widget)
        self.table_model.save_model()
        self.save_proxy_settings()

    def recursive_widget_saver(self, widget):
        """Writes out the data for this widget and all of its children, recursively."""
        if isinstance(widget, QtWidgets.QWidget):
            # See if it's one of ours:
            pref_path = widget.property("prefPath")
            pref_entry = widget.property("prefEntry")
            if pref_path and pref_entry:
                pref_path = pref_path.data()
                pref_entry = pref_entry.data()
                pref_access_string = f"User parameter:BaseApp/Preferences/{str(pref_path,'utf-8')}"
                pref = fci.FreeCAD.ParamGet(pref_access_string)
                if isinstance(widget, QtWidgets.QCheckBox):
                    checked = widget.isChecked()
                    pref.SetBool(str(pref_entry, "utf-8"), checked)
                elif isinstance(widget, QtWidgets.QRadioButton):
                    checked = widget.isChecked()
                    pref.SetBool(str(pref_entry, "utf-8"), checked)
                elif isinstance(widget, QtWidgets.QComboBox):
                    new_index = widget.currentIndex()
                    pref.SetInt(str(pref_entry, "utf-8"), new_index)
                elif isinstance(widget, QtWidgets.QTextEdit):
                    text = widget.toPlainText()
                    pref.SetString(str(pref_entry, "utf-8"), text)
                elif isinstance(widget, QtWidgets.QLineEdit):
                    text = widget.text()
                    pref.SetString(str(pref_entry, "utf-8"), text)
                elif widget.metaObject().className() == "Gui::PrefFileChooser":
                    filename = str(widget.property("fileName"))
                    pref.SetString(str(pref_entry, "utf-8"), filename)

        # Recurse over children
        if isinstance(widget, QtCore.QObject):
            for child in widget.children():
                self.recursive_widget_saver(child)

    def _proxy_type_from_ui(self) -> ProxyType:
        if self.form.proxyGroupBox.isChecked():
            if self.form.systemProxyButton.isChecked():
                return AddonManagerOptions.ProxyType.system
            return AddonManagerOptions.ProxyType.custom
        return AddonManagerOptions.ProxyType.none

    def save_proxy_settings(self):
        """Save the proxy settings -- the line edits are taken care of by the widgets, but the
        check state of the group box, and the selection state of the two buttons, must be manually
        determined and stored."""
        proxy_type = str(self._proxy_type_from_ui())
        host = self.form.proxyHostLineEdit.text() if proxy_type == "custom" else ""
        port = int(self.form.proxyPortLineEdit.text()) if proxy_type == "custom" else 8080

        if (
            fci.Preferences().get("proxy_type") != proxy_type
            or fci.Preferences().get("proxy_host") != host
            or fci.Preferences().get("proxy_port") != port
        ):
            fci.Preferences().set("proxy_type", proxy_type)
            fci.Preferences().set("proxy_host", self.form.proxyHostLineEdit.text())
            fci.Preferences().set("proxy_port", int(self.form.proxyPortLineEdit.text()))
            ForceReinitializeNetworkManager()

    def loadSettings(self):
        """Required function: called by the preferences dialog when it is launched,
        loads the preference data and assigns it to the widgets."""
        for widget in self.form.children():
            self.recursive_widget_loader(widget)
        self.table_model.load_model()

        proxy_type = AddonManagerOptions.ProxyType.from_string(fci.Preferences().get("proxy_type"))
        self.reconfigure_proxy_ui(proxy_type)
        self.fill_proxy_host_settings_from_preferences()

    def recursive_widget_loader(self, widget):
        """Loads the data for this widget and all of its children, recursively."""
        if isinstance(widget, QtWidgets.QWidget):
            # See if it's one of ours:
            pref_path = widget.property("prefPath")
            pref_entry = widget.property("prefEntry")
            if pref_path and pref_entry:
                pref_path = pref_path.data()
                pref_entry = pref_entry.data()
                pref_access_string = f"User parameter:BaseApp/Preferences/{str(pref_path,'utf-8')}"
                pref = fci.FreeCAD.ParamGet(pref_access_string)
                if isinstance(widget, QtWidgets.QCheckBox):
                    widget.setChecked(pref.GetBool(str(pref_entry, "utf-8")))
                elif isinstance(widget, QtWidgets.QRadioButton):
                    if pref.GetBool(str(pref_entry, "utf-8")):
                        widget.setChecked(True)
                elif isinstance(widget, QtWidgets.QComboBox):
                    new_index = pref.GetInt(str(pref_entry, "utf-8"))
                    widget.setCurrentIndex(new_index)
                elif isinstance(widget, QtWidgets.QTextEdit):
                    text = pref.GetString(str(pref_entry, "utf-8"))
                    widget.setText(text)
                elif isinstance(widget, QtWidgets.QLineEdit):
                    text = pref.GetString(str(pref_entry, "utf-8"))
                    widget.setText(text)
                elif widget.metaObject().className() == "Gui::PrefFileChooser":
                    filename = pref.GetString(str(pref_entry, "utf-8"))
                    widget.setProperty("fileName", filename)

        # Recurse over children
        if isinstance(widget, QtCore.QObject):
            for child in widget.children():
                self.recursive_widget_loader(child)

    def _add_custom_repo_clicked(self):
        """Callback: show the Add custom repo dialog"""
        dlg = CustomRepositoryDialog()
        url, branch = dlg.exec()
        if url and branch:
            self.table_model.appendData(url, branch)

    def _remove_custom_repo_clicked(self):
        """Callback: when the remove button is clicked, get the current selection and remove it."""
        item = self.form.customRepositoriesTableView.currentIndex()
        if not item.isValid():
            return
        row = item.row()
        self.table_model.removeRows(row, 1, QtCore.QModelIndex())

    def _row_double_clicked(self, item):
        """Edit the row that was double-clicked"""
        row = item.row()
        dlg = CustomRepositoryDialog()
        url_index = self.table_model.createIndex(row, 0)
        branch_index = self.table_model.createIndex(row, 1)
        dlg.dialog.urlLineEdit.setText(self.table_model.data(url_index))
        dlg.dialog.branchLineEdit.setText(self.table_model.data(branch_index))
        url, branch = dlg.exec()
        if url and branch:
            self.table_model.setData(url_index, url)
            self.table_model.setData(branch_index, branch)


class CustomRepoDataModel(QtCore.QAbstractTableModel):
    """The model for the custom repositories: wraps the underlying preference data and uses that
    as its main data store."""

    def __init__(self):
        super().__init__()
        pref_access_string = "User parameter:BaseApp/Preferences/Addons"
        self.pref = fci.FreeCAD.ParamGet(pref_access_string)
        self.model = []
        self.load_model()

    def load_model(self):
        """Load the data from the preference entry"""
        pref_entry: str = self.pref.GetString("CustomRepositories", "")

        # The entry is saved as a space- and newline-delimited text block: break it into its
        # constituent parts
        lines = pref_entry.split("\n")
        self.model = []
        for line in lines:
            if not line:
                continue
            split_data = line.split()
            if len(split_data) > 1:
                branch = split_data[1]
            else:
                branch = "master"
            url = split_data[0]
            self.model.append([url, branch])

    def save_model(self):
        """Save the data into a preferences entry"""
        entry = ""
        for row in self.model:
            entry += f"{row[0]} {row[1]}\n"
        self.pref.SetString("CustomRepositories", entry)

    def rowCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        """The number of rows"""
        if parent.isValid():
            return 0
        return len(self.model)

    def columnCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        """The number of columns (which is always 2)"""
        if parent.isValid():
            return 0
        return 2

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """The data at an index."""
        if role != QtCore.Qt.DisplayRole:
            return None
        row = index.row()
        column = index.column()
        if row > len(self.model):
            return None
        if column > 1:
            return None
        return self.model[row][column]

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        """Get the row and column header data."""
        if role != QtCore.Qt.DisplayRole:
            return None
        if orientation == QtCore.Qt.Vertical:
            return section + 1
        if section == 0:
            return translate(
                "AddonsInstaller",
                "Repository URL",
                "Preferences header for custom repositories",
            )
        if section == 1:
            return translate(
                "AddonsInstaller",
                "Branch name",
                "Preferences header for custom repositories",
            )
        return None

    def removeRows(
        self, row: int, count: int, parent: QtCore.QModelIndex | QtCore.QPersistentModelIndex
    ) -> bool:
        """Remove rows"""
        self.beginRemoveRows(parent, row, row + count - 1)
        for _ in range(count):
            self.model.pop(row)
        self.endRemoveRows()
        return True

    def insertRows(
        self, row: int, count: int, parent: QtCore.QModelIndex | QtCore.QPersistentModelIndex
    ) -> bool:
        """Insert blank rows"""
        self.beginInsertRows(parent, row, row + count - 1)
        for _ in range(count):
            self.model.insert(row, ["", ""])
        self.endInsertRows()
        return True

    def appendData(self, url, branch):
        """Append this url and branch to the end of the list"""
        row = self.rowCount()
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.model.append([url, branch])
        self.endInsertRows()

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        """Set the data at this index"""
        if role != QtCore.Qt.EditRole:
            return
        self.model[index.row()][index.column()] = value
        self.dataChanged.emit(index, index)


class CustomRepositoryDialog:
    """A dialog for setting up a custom repository, with branch information"""

    def __init__(self):
        self.dialog = fci.loadUi(
            os.path.join(os.path.dirname(__file__), "AddonManagerOptions_AddCustomRepository.ui")
        )
        self.dialog.setObjectName("AddonManager_AddCustomRepositoryDialog")

    def exec(self):
        """Run the dialog as a modal, and return either None or a tuple of (url,branch)"""
        result = self.dialog.exec()
        if result == QtWidgets.QDialog.Accepted:
            url = self.dialog.urlLineEdit.text()
            branch = self.dialog.branchLineEdit.text()
            return url, branch
        return None, None



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix write inpfile femelement geometry"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def write_femelement_geometry(f, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    f.write("\n{}\n".format(59 * "*"))
    f.write("** Sections\n")
    for matgeoset in ccxwriter.mat_geo_sets:
        if matgeoset["ccx_elset"]:
            elsetdef = "ELSET={}, ".format(matgeoset["ccx_elset_name"])
            material = "MATERIAL={}".format(matgeoset["mat_obj_name"])

            if "beamsection_obj" in matgeoset:  # beam mesh
                beamsec_obj = matgeoset["beamsection_obj"]
                beam_axis_m = matgeoset["beam_axis_m"]
                # in CalxuliX called the 1direction
                # see meshtools.get_beam_main_axis_m(beam_direction, defined_angle)
                section_nor = "{:.13G}, {:.13G}, {:.13G}\n".format(
                    beam_axis_m[0], beam_axis_m[1], beam_axis_m[2]
                )
                if ccxwriter.solver_obj.ExcludeBendingStiffness:
                    area = beamsec_obj.TrussArea.getValueAs("mm^2").Value
                    section_def = f"*SOLID SECTION, {elsetdef}{material}\n"
                    section_geo = f"{area:.13G}\n"
                else:
                    if beamsec_obj.SectionType == "Rectangular":
                        # see meshtools.get_beam_main_axis_m(beam_direction, defined_angle)
                        # the method get_beam_main_axis_m() which calculates the beam_axis_m vector
                        # unless rotated, this vector points towards +y axis
                        # doesn't follow 1,2-direction order of CalculiX
                        # ^ (n, 2-direction)
                        # |
                        # |
                        # .----> (m, 1-direction)
                        #
                        len_beam_axis_n = beamsec_obj.RectHeight.getValueAs("mm").Value
                        len_beam_axis_m = beamsec_obj.RectWidth.getValueAs("mm").Value
                        section_type = ", SECTION=RECT"
                        section_geo = f"{len_beam_axis_m:.13G},{len_beam_axis_n:.13G}\n"
                        section_def = f"*BEAM SECTION, {elsetdef}{material}{section_type}\n"
                    elif beamsec_obj.SectionType == "Circular":
                        diameter = beamsec_obj.CircDiameter.getValueAs("mm").Value
                        section_type = ", SECTION=CIRC"
                        section_geo = f"{diameter:.13G}\n"
                        section_def = f"*BEAM SECTION, {elsetdef}{material}{section_type}\n"
                    elif beamsec_obj.SectionType == "Elliptical":
                        axis1 = beamsec_obj.Axis1Length.getValueAs("mm").Value
                        axis2 = beamsec_obj.Axis2Length.getValueAs("mm").Value
                        section_type = ", SECTION=CIRC"
                        section_geo = f"{axis1:.13G},{axis2:.13G}\n"
                        section_def = f"*BEAM SECTION, {elsetdef}{material}{section_type}\n"
                    elif beamsec_obj.SectionType == "Pipe":
                        radius = 0.5 * beamsec_obj.PipeDiameter.getValueAs("mm").Value
                        thickness = beamsec_obj.PipeThickness.getValueAs("mm").Value
                        section_type = ", SECTION=PIPE"
                        section_geo = f"{radius:.13G},{thickness:.13G}\n"
                        section_def = f"*BEAM SECTION, {elsetdef}{material}{section_type}\n"
                    elif beamsec_obj.SectionType == "Box":
                        box_width = beamsec_obj.BoxWidth.getValueAs("mm").Value
                        box_height = beamsec_obj.BoxHeight.getValueAs("mm").Value
                        box_t1 = beamsec_obj.BoxT1.getValueAs("mm").Value
                        box_t2 = beamsec_obj.BoxT2.getValueAs("mm").Value
                        box_t3 = beamsec_obj.BoxT3.getValueAs("mm").Value
                        box_t4 = beamsec_obj.BoxT4.getValueAs("mm").Value
                        section_type = ", SECTION=BOX"
                        section_geo = f"{box_width:.13G},{box_height:.13G},{box_t1:.13G},{box_t2:.13G},{box_t3:.13G},{box_t4:.13G}\n"
                        section_def = f"*BEAM SECTION, {elsetdef}{material}{section_type}\n"

                f.write(section_def)
                f.write(section_geo)
                if not ccxwriter.solver_obj.ExcludeBendingStiffness:
                    f.write(section_nor)
            elif "fluidsection_obj" in matgeoset:  # fluid mesh
                fluidsec_obj = matgeoset["fluidsection_obj"]
                if fluidsec_obj.SectionType == "Liquid":
                    section_type = fluidsec_obj.LiquidSectionType
                    if (section_type == "PIPE INLET") or (section_type == "PIPE OUTLET"):
                        section_type = "PIPE INOUT"
                    section_def = "*FLUID SECTION, {}TYPE={}, {}\n".format(
                        elsetdef, section_type, material
                    )
                    section_geo = liquid_section_def(fluidsec_obj, section_type)
                """
                # deactivate as it would result in section_def and section_geo not defined
                # deactivated in the App and Gui object and thus in the task panel as well
                elif fluidsec_obj.SectionType == "Gas":
                    section_type = fluidsec_obj.GasSectionType
                elif fluidsec_obj.SectionType == "Open Channel":
                    section_type = fluidsec_obj.ChannelSectionType
                """
                f.write(section_def)
                f.write(section_geo)
            elif "shellthickness_obj" in matgeoset:  # shell mesh
                shellth_obj = matgeoset["shellthickness_obj"]
                if ccxwriter.solver_obj.ModelSpace == "3D":
                    offset = shellth_obj.Offset
                    if ccxwriter.solver_obj.ExcludeBendingStiffness:
                        section_def = (
                            f"*MEMBRANE SECTION, {elsetdef}{material}, OFFSET={offset:.13G}\n"
                        )
                    else:
                        section_def = (
                            f"*SHELL SECTION, {elsetdef}{material}, OFFSET={offset:.13G}\n"
                        )
                else:
                    section_def = f"*SOLID SECTION, {elsetdef}{material}\n"
                thickness = shellth_obj.Thickness.getValueAs("mm").Value
                section_geo = f"{thickness:.13G}\n"
                f.write(section_def)
                f.write(section_geo)
            else:  # solid mesh
                section_def = f"*SOLID SECTION, {elsetdef}{material}\n"
                f.write(section_def)


# ************************************************************************************************
# Helpers
def liquid_section_def(obj, section_type):
    if section_type == "PIPE MANNING":
        manning_area = obj.ManningArea.getValueAs("mm^2").Value
        manning_radius = obj.ManningRadius.getValueAs("mm").Value
        manning_coefficient = obj.ManningCoefficient
        section_geo = "{:.13G},{:.13G},{:.13G}\n".format(
            manning_area, manning_radius, manning_coefficient
        )
        return section_geo
    elif section_type == "PIPE ENLARGEMENT":
        enlarge_area1 = obj.EnlargeArea1.getValueAs("mm^2").Value
        enlarge_area2 = obj.EnlargeArea2.getValueAs("mm^2").Value
        section_geo = f"{enlarge_area1:.13G},{enlarge_area2:.13G}\n"
        return section_geo
    elif section_type == "PIPE CONTRACTION":
        contract_area1 = obj.ContractArea1.getValueAs("mm^2").Value
        contract_area2 = obj.ContractArea2.getValueAs("mm^2").Value
        section_geo = f"{contract_area1:.13G},{contract_area2:.13G}\n"
        return section_geo
    elif section_type == "PIPE ENTRANCE":
        entrance_pipe_area = obj.EntrancePipeArea.getValueAs("mm^2").Value
        entrance_area = obj.EntranceArea.getValueAs("mm^2").Value
        section_geo = f"{entrance_pipe_area:.13G},{entrance_area:.13G}\n"
        return section_geo
    elif section_type == "PIPE DIAPHRAGM":
        diaphragm_pipe_area = obj.DiaphragmPipeArea.getValueAs("mm^2").Value
        diaphragm_area = obj.DiaphragmArea.getValueAs("mm^2").Value
        section_geo = f"{diaphragm_pipe_area:.13G},{diaphragm_area:.13G}\n"
        return section_geo
    elif section_type == "PIPE BEND":
        bend_pipe_area = obj.BendPipeArea.getValueAs("mm^2").Value
        bend_radius_diameter = obj.BendRadiusDiameter
        bend_angle = obj.BendAngle
        bend_loss_coefficient = obj.BendLossCoefficient
        section_geo = "{:.13G},{:.13G},{:.13G},{:.13G}\n".format(
            bend_pipe_area, bend_radius_diameter, bend_angle, bend_loss_coefficient
        )
        return section_geo
    elif section_type == "PIPE GATE VALVE":
        gatevalve_pipe_area = obj.GateValvePipeArea.getValueAs("mm^2").Value
        gatevalve_closing_coeff = obj.GateValveClosingCoeff
        section_geo = f"{gatevalve_pipe_area:.13G},{gatevalve_closing_coeff:.13G}\n"
        return section_geo
    elif section_type == "PIPE WHITE-COLEBROOK":
        colebrooke_area = obj.ColebrookeArea.getValueAs("mm^2").Value
        colebrooke_diameter = 2 * obj.ColebrookeRadius.getValueAs("mm")
        colebrooke_grain_diameter = obj.ColebrookeGrainDiameter.getValueAs("mm")
        colebrooke_form_factor = obj.ColebrookeFormFactor
        section_geo = "{:.13G},{:.13G},{},{:.13G},{:.13G}\n".format(
            colebrooke_area,
            colebrooke_diameter,
            "-1",
            colebrooke_grain_diameter,
            colebrooke_form_factor,
        )
        return section_geo
    elif section_type == "LIQUID PUMP":
        section_geo = ""
        for i in range(len(obj.PumpFlowRate)):
            flow_rate = obj.PumpFlowRate[i]
            top = obj.PumpHeadLoss[i]
            section_geo = f"{section_geo + flow_rate:.13G},{top:.13G},\n"
        section_geo = f"{section_geo}\n"
        return section_geo
    else:
        return ""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


from PySide.QtCore import QT_TRANSLATE_NOOP
from threading import Thread, Lock
import FreeCAD
import FreeCADGui
import Mesh
import Path
import PathScripts
import PathScripts.PathUtils as PathUtils
import Path.Post.Command as PathPost
import camotics
import io
import json
import queue
import subprocess

from PySide import QtCore, QtGui

__title__ = "Camotics Simulator"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Task panel for Camotics Simulation"

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class CAMoticsUI:
    def __init__(self, simulation):
        # this will create a Qt widget from our ui file
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/TaskPathCamoticsSim.ui")
        self.simulation = simulation
        self.initializeUI()
        self.lock = False

    def initializeUI(self):
        self.form.timeSlider.sliderReleased.connect(
            lambda: self.simulation.execute(self.form.timeSlider.value())
        )
        self.form.progressBar.reset()
        self.form.timeSlider.setEnabled = False
        self.form.btnLaunchCamotics.clicked.connect(self.launchCamotics)
        self.form.btnMakeFile.clicked.connect(self.makeCamoticsFile)
        self.simulation.progressUpdate.connect(self.calculating)
        self.simulation.statusChange.connect(self.updateStatus)
        self.form.txtStatus.setText(translate("Path", "Drag Slider to Simulate"))

    def launchCamotics(self):
        filename = self.makeCamoticsFile()
        subprocess.Popen(["camotics", filename])

    def makeCamoticsFile(self):
        Path.Log.track()
        filename = QtGui.QFileDialog.getSaveFileName(
            self.form,
            translate("Path", "Save Project As"),
            "",
            translate("Path", "CAMotics Project (*.camotics)"),
        )[0]
        if filename:
            if not filename.endswith(".camotics"):
                filename += ".camotics"

            text = self.simulation.buildproject()
            try:
                with open(filename, "w") as outputfile:
                    outputfile.write(text)
            except IOError:
                QtGui.QMessageBox.information(
                    self, translate("Path", "Unable to open file: {}".format(filename))
                )

        return filename

    def accept(self):
        self.simulation.accept()
        FreeCADGui.Control.closeDialog()

    def reject(self):
        self.simulation.cancel()
        if self.simulation.simmesh is not None:
            FreeCAD.ActiveDocument.removeObject(self.simulation.simmesh.Name)
        FreeCADGui.Control.closeDialog()

    def setRunTime(self, duration):
        self.form.timeSlider.setMinimum(0)
        self.form.timeSlider.setMaximum(duration)

    def calculating(self, progress=0.0):
        self.form.timeSlider.setEnabled = progress == 1.0
        self.form.progressBar.setValue(int(progress * 100))

    def updateStatus(self, status):
        self.form.txtStatus.setText(status)


class CamoticsSimulation(QtCore.QObject):

    SIM = camotics.Simulation()
    q = queue.Queue()
    progressUpdate = QtCore.Signal(object)
    statusChange = QtCore.Signal(object)
    simmesh = None
    filenames = []

    SHAPEMAP = {
        "ballend": "Ballnose",
        "endmill": "Cylindrical",
        "taperedballnose": "Ballnose",
        "v-bit": "Conical",
        "chamfer": "Snubnose",
    }

    def worker(self, lock):
        while True:
            item = self.q.get()
            Path.Log.debug("worker processing: {}".format(item))
            with lock:
                if item["TYPE"] == "STATUS":
                    self.statusChange.emit(item["VALUE"])
                    if item["VALUE"] == "DONE":
                        self.SIM.wait()
                        surface = self.SIM.get_surface("binary")
                        self.SIM.wait()
                        self.addMesh(surface)
                elif item["TYPE"] == "PROGRESS":
                    self.progressUpdate.emit(item["VALUE"])
            self.q.task_done()

    def __init__(self):
        super().__init__()  # needed for QT signals
        lock = Lock()
        Thread(target=self.worker, daemon=True, args=(lock,)).start()

    def callback(self, status, progress):
        self.q.put({"TYPE": "PROGRESS", "VALUE": progress})
        self.q.put({"TYPE": "STATUS", "VALUE": status})

    def isDone(self, success):
        self.q.put({"TYPE": "STATUS", "VALUE": "DONE"})

    def addMesh(self, surface):
        """takes a binary stl and adds a Mesh to the current document"""

        if self.simmesh is None:
            self.simmesh = FreeCAD.ActiveDocument.addObject("Mesh::Feature", "Camotics")
        buffer = io.BytesIO()
        buffer.write(surface)
        buffer.seek(0)
        mesh = Mesh.Mesh()
        mesh.read(buffer, "STL")
        self.simmesh.Mesh = mesh
        # Mesh.show(mesh)

    def Activate(self):
        self.taskForm = CAMoticsUI(self)
        FreeCADGui.Control.showDialog(self.taskForm)
        self.job = FreeCADGui.Selection.getSelectionEx()[0].Object
        self.SIM.set_metric()
        self.SIM.set_resolution("high")

        bb = self.job.Stock.Shape.BoundBox
        self.SIM.set_workpiece(min=(bb.XMin, bb.YMin, bb.ZMin), max=(bb.XMax, bb.YMax, bb.ZMax))

        for t in self.job.Tools.Group:
            self.SIM.set_tool(
                t.ToolNumber,
                metric=True,
                shape=self.SHAPEMAP.get(PathUtils.getToolShapeName(t.Tool), "Cylindrical"),
                length=t.Tool.Length.Value,
                diameter=t.Tool.Diameter.Value,
            )

        postlist = PathPost.buildPostList(self.job)
        Path.Log.track(postlist)
        # self.filenames = [PathPost.resolveFileName(self.job)]

        success = True

        finalgcode = ""
        for idx, section in enumerate(postlist):
            partname = section[0]
            sublist = section[1]

            result, gcode, name = PathPost.CommandPathPost().exportObjectsWith(
                sublist,
                partname,
                self.job,
                idx,
                extraargs="--no-show-editor",
            )
            self.filenames.append(name)
            Path.Log.track(result, gcode, name)

            if result is None:
                success = False
            else:
                finalgcode += gcode

        if not success:
            return

        self.SIM.compute_path(finalgcode)
        self.SIM.wait()

        tot = sum([step["time"] for step in self.SIM.get_path()])
        Path.Log.debug("sim time: {}".format(tot))
        self.taskForm.setRunTime(tot)

    def execute(self, timeIndex):
        Path.Log.track()
        self.SIM.start(self.callback, time=timeIndex, done=self.isDone)

    def accept(self):
        pass

    def cancel(self):
        pass

    def buildproject(self):  # , files=[]):
        Path.Log.track()

        job = self.job

        tooltemplate = {
            "units": "metric",
            "shape": "cylindrical",
            "length": 10,
            "diameter": 3.125,
            "description": "",
        }

        workpiecetemplate = {
            "automatic": False,
            "margin": 0,
            "bounds": {"min": [0, 0, 0], "max": [0, 0, 0]},
        }

        camoticstemplate = {
            "units": "metric",
            "resolution-mode": "medium",
            "resolution": 1,
            "tools": {},
            "workpiece": {},
            "files": [],
        }

        unitstring = "imperial" if FreeCAD.Units.getSchema() in [2, 3, 5, 7] else "metric"

        camoticstemplate["units"] = unitstring
        camoticstemplate["resolution-mode"] = "medium"
        camoticstemplate["resolution"] = 1

        toollist = {}
        for t in job.Tools.Group:
            toolitem = tooltemplate.copy()
            toolitem["units"] = unitstring
            if hasattr(t.Tool, "Camotics"):
                toolitem["shape"] = t.Tool.Camotics
            else:
                toolitem["shape"] = self.SHAPEMAP.get(
                    PathUtils.getToolShapeName(t.Tool), "Cylindrical"
                )

            toolitem["length"] = t.Tool.Length.Value
            toolitem["diameter"] = t.Tool.Diameter.Value
            toolitem["description"] = t.Label
            toollist[t.ToolNumber] = toolitem

        camoticstemplate["tools"] = toollist

        bb = job.Stock.Shape.BoundBox

        workpiecetemplate["bounds"]["min"] = [bb.XMin, bb.YMin, bb.ZMin]
        workpiecetemplate["bounds"]["max"] = [bb.XMax, bb.YMax, bb.ZMax]
        camoticstemplate["workpiece"] = workpiecetemplate

        camoticstemplate["files"] = self.filenames  # files

        return json.dumps(camoticstemplate, indent=2)


class CommandCamoticsSimulate:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Camotics",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Camotics", "CAMotics"),
            "Accel": "P, C",
            "ToolTip": QT_TRANSLATE_NOOP("CAM_Camotics", "Simulates using CAMotics"),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        if bool(FreeCADGui.Selection.getSelection()) is False:
            return False
        try:
            job = FreeCADGui.Selection.getSelectionEx()[0].Object
            return isinstance(job.Proxy, Path.Main.Job.ObjectJob)
        except:
            return False

    def Activated(self):
        pathSimulation = CamoticsSimulation()
        pathSimulation.Activate()


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("CAM_Camotics", CommandCamoticsSimulate())


FreeCAD.Console.PrintLog("Loading PathCamoticsSimulateGui… done\n")



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint heatflux"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def get_analysis_types():
    return ["thermomech"]


def get_sets_name():
    return "constraints_heatflux_element_face_heatflux"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def write_meshdata_constraint(f, femobj, heatflux_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    if heatflux_obj.ConstraintType == "Convection":
        heatflux_key_word = "FILM"
        heatflux_facetype = "F"
        heatflux_facesubtype = ""
        heatflux_values = "{:.13G},{:.13G}".format(
            heatflux_obj.AmbientTemp.getValueAs("K").Value,
            heatflux_obj.FilmCoef.getValueAs("t/s^3/K").Value,
        )

    elif heatflux_obj.ConstraintType == "Radiation":
        heatflux_facetype = "R"
        if heatflux_obj.CavityRadiation:
            heatflux_key_word = f"RADIATE, CAVITY={heatflux_obj.CavityName}"
            heatflux_facesubtype = "CR"
        else:
            heatflux_key_word = "RADIATE"
            heatflux_facesubtype = ""
        heatflux_values = "{:.13G},{:.13G}".format(
            heatflux_obj.AmbientTemp.getValueAs("K").Value, heatflux_obj.Emissivity
        )

    elif heatflux_obj.ConstraintType == "DFlux":
        heatflux_key_word = "DFLUX"
        heatflux_facetype = "S"
        heatflux_facesubtype = ""
        heatflux_values = "{:.13G}".format(heatflux_obj.DFlux.getValueAs("t/s^3").Value)

    else:
        return

    if heatflux_obj.EnableAmplitude:
        heatflux_amplitude = f", AMPLITUDE={heatflux_obj.Name}"
    else:
        heatflux_amplitude = ""

    f.write(f"*{heatflux_key_word}{heatflux_amplitude}\n")
    for feat, surf, is_sub_el in femobj["HeatFluxFaces"]:
        f.write("** {0.Name}.{1[0]}\n".format(*feat))
        for face, fno in surf:
            f.write(f"{face},{heatflux_facetype}{fno}{heatflux_facesubtype},{heatflux_values}\n")



# ***************************************************************************
# *   Copyright (c) 2022 Ajinkya Dahale <dahale.a.p@gmail.com>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint rigid body"
__author__ = "Ajinkya Dahale"
__url__ = "https://www.freecad.org"


def get_analysis_types():
    return "all"  # write for all analysis types


def get_sets_name():
    return "constraints_rigidbody_node_sets"


def get_constraint_title():
    return "Rigid Body Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, rb_obj, ccxwriter):

    f.write("*NSET,NSET=" + rb_obj.Name + "\n")
    for n in femobj["Nodes"]:
        f.write(f"{n},\n")


def write_constraint(f, femobj, rb_obj, ccxwriter):

    rb_obj_idx = ccxwriter.analysis.Group.index(rb_obj)
    node_count = ccxwriter.mesh_object.FemMesh.NodeCount
    # factor 2 is to prevent conflict with other rigid body constraint
    ref_node_idx = node_count + 2 * rb_obj_idx + 1
    rot_node_idx = node_count + 2 * rb_obj_idx + 2

    f.write("*NODE\n")
    f.write("{},{},{},{}\n".format(ref_node_idx, *rb_obj.ReferenceNode))
    f.write("{},{},{},{}\n".format(rot_node_idx, *rb_obj.ReferenceNode))

    f.write(f"*NSET,NSET={rb_obj.Name}_RefNode\n")
    f.write(f"{ref_node_idx},\n")
    f.write(f"*NSET,NSET={rb_obj.Name}_RotNode\n")
    f.write(f"{rot_node_idx},\n")

    kw_line = f"*RIGID BODY, NSET={rb_obj.Name}, REF NODE={ref_node_idx}, ROT NODE={rot_node_idx}"

    f.write(kw_line + "\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import Path

__title__ = "Setup Sheet for a Job."
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Prototype objects to allow extraction of setup sheet values and editing."


Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
# Path.Log.trackModule(Path.Log.thisModule())


class Property(object):
    """Base class for all prototype properties"""

    def __init__(self, name, propType, category, info):
        self.name = name
        self.propType = propType
        self.category = category
        self.info = info
        self.editorMode = 0
        self.value = None

    def setValue(self, value):
        self.value = value

    def getValue(self):
        return self.value

    def setEditorMode(self, mode):
        self.editorMode = mode

    def displayString(self):
        if self.value is None:
            t = self.typeString()
            p = "an" if t[0] in ["A", "E", "I", "O", "U"] else "a"
            return "%s %s" % (p, t)
        return self.value

    def typeString(self):
        return "Property"

    def setupProperty(self, obj, name, category, value):
        created = False
        if not hasattr(obj, name):
            Path.Log.track("add", obj.Name, name, self.propType)
            obj.addProperty(self.propType, name, category, self.info)
            self.initProperty(obj, name)
            created = True
        Path.Log.track("set", obj.Name, name, value, type(value))
        setattr(obj, name, value)
        return created

    def initProperty(self, obj, name):
        pass

    def setValueFromString(self, string):
        self.setValue(self.valueFromString(string))

    def valueFromString(self, string):
        return string


class PropertyEnumeration(Property):
    def typeString(self):
        return "Enumeration"

    def setValue(self, value):
        if list == type(value):
            self.enums = value
        else:
            super(PropertyEnumeration, self).setValue(value)

    def getEnumValues(self):
        return self.enums

    def initProperty(self, obj, name):
        setattr(obj, name, self.enums)


class PropertyQuantity(Property):
    def displayString(self):
        if self.value is None:
            return Property.displayString(self)
        return self.value.getUserPreferred()[0]


class PropertyAngle(PropertyQuantity):
    def typeString(self):
        return "Angle"


class PropertyDistance(PropertyQuantity):
    def typeString(self):
        return "Distance"


class PropertyLength(PropertyQuantity):
    def typeString(self):
        return "Length"


class PropertyFloat(Property):
    def typeString(self):
        return "Float"

    def valueFromString(self, string):
        try:
            return float(string)
        except ValueError:
            Path.Log.error(f"{self.category}.{self.name} [{self.propType}] : '{string}'")
            raise


class PropertyInteger(Property):
    def typeString(self):
        return "Integer"

    def valueFromString(self, string):
        try:
            return int(string)
        except ValueError:
            Path.Log.error(f"{self.category}.{self.name} [{self.propType}] : '{string}'")
            raise


class PropertyPercent(PropertyInteger):
    def typeString(self):
        return "Percent"


class PropertyBool(Property):
    def typeString(self):
        return "Bool"

    def valueFromString(self, string):
        try:
            return bool(string)
        except ValueError:
            Path.Log.error(f"{self.category}.{self.name} [{self.propType}] : '{string}'")
            raise


class PropertyString(Property):
    def typeString(self):
        return "String"


class PropertyMap(Property):
    def typeString(self):
        return "Map"

    def displayString(self, value):
        return str(value)


class OpPrototype(object):

    PropertyType = {
        "App::PropertyAngle": PropertyAngle,
        "App::PropertyBool": PropertyBool,
        "App::PropertyDistance": PropertyDistance,
        "App::PropertyEnumeration": PropertyEnumeration,
        "App::PropertyFile": PropertyString,
        "App::PropertyFloat": PropertyFloat,
        "App::PropertyFloatConstraint": Property,
        "App::PropertyFloatList": Property,
        "App::PropertyInteger": PropertyInteger,
        "App::PropertyIntegerList": PropertyInteger,
        "App::PropertyLength": PropertyLength,
        "App::PropertyLink": Property,
        "App::PropertyLinkList": Property,
        "App::PropertyLinkSubListGlobal": Property,
        "App::PropertyMap": PropertyMap,
        "App::PropertyPercent": PropertyPercent,
        "App::PropertyString": PropertyString,
        "App::PropertyStringList": Property,
        "App::PropertyVectorDistance": Property,
        "App::PropertyVectorList": Property,
        "Part::PropertyPartShape": Property,
        "App::PropertyPythonObject": PropertyString,
    }

    def __init__(self, name):
        self.Label = name
        self.properties = {}
        self.DoNotSetDefaultValues = True
        self.Proxy = None

    def __setattr__(self, name, val):
        if name in ["Label", "DoNotSetDefaultValues", "properties", "Proxy"]:
            if name == "Proxy":
                val = None  # make sure the proxy is never set
            return super(OpPrototype, self).__setattr__(name, val)
        self.properties[name].setValue(val)

    def addProperty(self, typeString, name, category, info=None):
        prop = self.PropertyType[typeString](name, typeString, category, info)
        self.properties[name] = prop
        return self

    def setEditorMode(self, name, mode):
        self.properties[name].setEditorMode(mode)

    def setPropertyStatus(self, name, status):
        pass

    def getProperty(self, name):
        return self.properties[name]

    def setupProperties(self, setup):
        return [p for p in self.properties if p.name in setup]



# **************************************************************************
#   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
#                                                                         *
#   This file is part of the FreeCAD CAx development system.              *
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU Lesser General Public License (LGPL)    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#   for detail see the LICENCE text file.                                 *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful,            *
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#   GNU Library General Public License for more details.                  *
#                                                                         *
#   You should have received a copy of the GNU Library General Public     *
#   License along with FreeCAD; if not, write to the Free Software        *
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#   USA                                                                   *
# **************************************************************************

# ---------------------------------------------------------------------------
# define the test cases to test the FreeCAD PartDesign module
# ---------------------------------------------------------------------------

# datum tools
from PartDesignTests.TestDatum import TestDatumPoint, TestDatumLine, TestDatumPlane
from PartDesignTests.TestShapeBinder import TestShapeBinder
from PartDesignTests.TestShapeBinder import TestSubShapeBinder

# additive/subtractive features & primitives
from PartDesignTests.TestPad import TestPad
from PartDesignTests.TestPocket import TestPocket
from PartDesignTests.TestHelix import TestHelix
from PartDesignTests.TestHole import TestHole
from PartDesignTests.TestRevolve import TestRevolve
from PartDesignTests.TestPipe import TestPipe
from PartDesignTests.TestLoft import TestLoft
from PartDesignTests.TestPrimitive import TestPrimitive
from PartDesignTests.TestHelix import TestHelix

# transformations and boolean
from PartDesignTests.TestMirrored import TestMirrored
from PartDesignTests.TestLinearPattern import TestLinearPattern
from PartDesignTests.TestPolarPattern import TestPolarPattern
from PartDesignTests.TestMultiTransform import TestMultiTransform
from PartDesignTests.TestBoolean import TestBoolean

# dressup features
from PartDesignTests.TestFillet import TestFillet
from PartDesignTests.TestChamfer import TestChamfer
from PartDesignTests.TestDraft import TestDraft
from PartDesignTests.TestThickness import TestThickness

# extras
from PartDesignTests.TestInvoluteGear import TestInvoluteGear
from PartDesignTests.TestSketch import TestSketch

# Topological naming problem
from PartDesignTests.TestTopologicalNamingProblem import TestTopologicalNamingProblem



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Block object."""
## @package block
# \ingroup draftobjects
# \brief Provides the object code for the Block object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

from draftobjects.base import DraftObject
from draftutils import gui_utils


class Block(DraftObject):
    """The Block object"""

    def __init__(self, obj):
        super().__init__(obj, "Block")

        _tip = QT_TRANSLATE_NOOP("App::Property", "The components of this block")
        obj.addProperty("App::PropertyLinkList", "Components", "Draft", _tip, locked=True)

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_base", vp_class="ViewProviderDraftPart", format=False
        )

    def execute(self, obj):
        if self.props_changed_placement_only(obj):
            obj.positionBySupport()
            self.props_changed_clear()
            return

        import Part

        plm = obj.Placement
        shps = []
        for c in obj.Components:
            shps.append(c.Shape)
        if shps:
            shape = Part.makeCompound(shps)
            obj.Shape = shape
        obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()

    def onChanged(self, obj, prop):
        self.props_changed_store(prop)


# Alias for compatibility with v0.18 and earlier
_Block = Block

## @}
